


def prune_leaves(tree, f):
    while tree.count_terminals() > 0:
        altered = False
        for a in tree.get_terminals():
            if f(a):
                tree.prune(a)
                altered = True
        if not altered:
            break


def align_cn_tree(tree, adata):
    prune_leaves(tree, lambda a: a.name not in adata.obs.index)

    cell_ids = []
    for a in tree.get_terminals():
        cell_ids.append(a.name)

    adata = adata[cell_ids]

    return tree, adata



import anndata as ad
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import pandas as pd
import Bio.Phylo

from anndata import AnnData

from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.patches import Patch
import matplotlib.cm

import scgenome.cnplot
import scgenome.refgenome
from . import cn_colors


def plot_cell_cn_matrix(
        adata: AnnData,
        layer_name='state',
        cell_order_fields=(),
        ax=None,
        raw=False,
        max_cn=13,
        vmin=None,
        vmax=None,
        show_cell_ids=False):
    """ Plot a copy number matrix

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data to plot, None for X, by default 'state'
    cell_order_fields : list, optional
        columns of obs on which to sort cells, by default None
    ax : matplotlib.axes.Axes, optional
        existing axis to plot into, by default None
    raw : bool, optional
        raw plotting, no integer color map, by default False
    max_cn : int, optional
        clip cn at max value, raw=False only, by default 13
    vmin, vmax : float, optional
        for raw=True, vmin and vmax define the data range that the colormap covers, see `matplotlib.pyplot.imshow`
    show_cell_ids : bool, optional
        show cell ids on heatmap axis, by default False

    Returns
    -------
    dict
        Dictionary of plot and data elements

    Examples
    -------

    .. plot::
        :context: close-figs

        import scgenome
        adata = scgenome.datasets.OV2295_HMMCopy_reduced()
        scgenome.pl.plot_cell_cn_matrix(adata)

    """

    if ax is None:
        ax = plt.gca()

    # Order the chromosomes
    chr_start = adata.var.reset_index().merge(scgenome.refgenome.info.chromosome_info[['chr', 'chr_index']], how='left')
    if chr_start['chr_index'].isnull().any():
        chromosomes = adata.var['chr'].astype(str).values
        raise ValueError(f'mismatching chromosomes {chromosomes} and {scgenome.refgenome.info.chromosomes}')
    chr_start = chr_start[['start', 'chr_index']].values
    genome_ordering = np.lexsort(chr_start.transpose())

    # Order the cells if requested
    if len(cell_order_fields) > 0:
        cell_order_fields = reversed(list(cell_order_fields))
        cell_order_values = adata.obs[cell_order_fields].values.transpose()
        cell_ordering = np.lexsort(cell_order_values)

    else:
        cell_ordering = range(adata.shape[0])

    adata = adata[cell_ordering, genome_ordering]

    if layer_name is not None:
        X = adata.layers[layer_name].copy()
    else:
        X = adata.X.copy()

    X = np.nan_to_num(X, nan=0)

    if not raw and max_cn is not None:
        X[X > max_cn] = max_cn

    if not raw:
        X_colors = cn_colors.map_cn_colors(X)
        im = ax.imshow(X_colors, aspect='auto', interpolation='none', vmin=vmin, vmax=vmax)

    else:
        cmap = matplotlib.cm.get_cmap('viridis')
        im = ax.imshow(X, aspect='auto', cmap=cmap, interpolation='none', vmin=vmin, vmax=vmax)

    mat_chrom_idxs = chr_start[genome_ordering][:, 1]
    chrom_boundaries = np.array([0] + list(np.where(mat_chrom_idxs[1:] != mat_chrom_idxs[:-1])[0]) + [mat_chrom_idxs.shape[0] - 1])
    chrom_sizes = chrom_boundaries[1:] - chrom_boundaries[:-1]
    chrom_mids = chrom_boundaries[:-1] + chrom_sizes / 2
    ordered_mat_chrom_idxs = mat_chrom_idxs[np.where(np.array([1] + list(np.diff(mat_chrom_idxs))) != 0)]
    chrom_names = np.array(scgenome.refgenome.info.plot_chromosomes)[ordered_mat_chrom_idxs]

    ax.set_xticks(chrom_mids)
    ax.set_xticklabels(chrom_names, fontsize='6')
    ax.set_xlabel('chromosome', fontsize=8)

    if show_cell_ids:
        ax.set(yticks=range(len(adata.obs.index)))
        ax.set(yticklabels=adata.obs.index.values)
        ax.tick_params(axis='y', labelrotation=0)
    else:
        ax.set(yticks=[])
        ax.set(yticklabels=[])

    for val in chrom_boundaries[:-1]:
        ax.axvline(x=val, linewidth=1, color='black', zorder=100)

    return {
        'ax': ax,
        'im': im,
        'adata': adata,
    }


def map_catagorigal_colors(values, cmap_name=None):
    levels = np.unique(values)
    n_levels = len(levels)

    if cmap_name is None:
        if n_levels <= 10:
            cmap_name = 'tab10'
        elif n_levels <= 20:
            cmap_name = 'tab20'
        else:
            cmap_name = 'hsv'

    cmap = matplotlib.cm.get_cmap(cmap_name)

    level_colors = dict(zip(levels, cmap(np.linspace(0, 1, n_levels))))

    value_colors = np.zeros(values.shape + (4,))
    for l, c in level_colors.items():
        value_colors[values == l, :] = c

    return level_colors, value_colors


def _plot_categorical_annotation(values, ax, ax_legend, title, horizontal=False):
    level_colors, value_colors = map_catagorigal_colors(values)

    im = ax.imshow(value_colors, aspect='auto', interpolation='none')

    levels = []
    patches = []
    for s, h in level_colors.items():
        levels.append(s)
        patches.append(Patch(facecolor=h, edgecolor=h))
    ncol = min(3, int(len(levels)**(1/2)))
    legend = ax_legend.legend(patches, levels, ncol=ncol,
        frameon=True, loc=2, bbox_to_anchor=(0., 1.),
        facecolor='white', edgecolor='white', fontsize='4',
        title=title, title_fontsize='6')

    ax.grid(False)
    if horizontal:
        ax.set_yticks([0.], [title], rotation=0, fontsize='6')
        ax.tick_params(axis='x', left=False, right=False)
    else:
        ax.set_xticks([0.], [title], rotation=90, fontsize='6')
        ax.tick_params(axis='y', left=False, right=False)

    annotation_info = {}
    annotation_info['ax'] = ax
    annotation_info['im'] = im
    annotation_info['ax_legend'] = ax_legend
    annotation_info['legend'] = legend
    annotation_info['level_colors'] = level_colors
    annotation_info['value_colors'] = value_colors

    return annotation_info


def _plot_continuous_legend(ax_legend, im, title):
    ax_legend.grid(False)
    ax_legend.set_xticks([])
    ax_legend.set_yticks([])

    axins = ax_legend.inset_axes([0.5, 0.1, 0.05, 0.8])

    cbar = plt.colorbar(im, cax=axins)
    axins.set_title(title, fontsize='6')
    cbar.ax.tick_params(labelsize='4')

    annotation_info = {}
    annotation_info['ax_legend'] = ax_legend
    annotation_info['axins'] = axins
    annotation_info['cbar'] = cbar

    return annotation_info


def _plot_continuous_annotation(values, ax, ax_legend, title, horizontal=False):

    im = ax.imshow(values, aspect='auto', interpolation='none', cmap='Reds')

    ax.grid(False)
    if horizontal:
        ax.set_yticks([0.], [title], rotation=0, fontsize='6')
        ax.tick_params(axis='x', left=False, right=False)
    else:
        ax.set_xticks([0.], [title], rotation=90, fontsize='6')
        ax.tick_params(axis='y', left=False, right=False)

    annotation_info = _plot_continuous_legend(ax_legend, im, title)

    annotation_info['ax'] = ax
    annotation_info['im'] = im

    return annotation_info


def plot_cell_cn_matrix_fig(
        adata: AnnData,
        layer_name='state',
        tree=None,
        cell_order_fields=None,
        annotation_fields=None,
        var_annotation_fields=None,
        fig=None,
        raw=False,
        vmin=None,
        vmax=None,
        max_cn=13,
        show_cell_ids=False,
        show_subsets=False):
    """ Plot a copy number matrix

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data to plot, None for X, by default 'state'
    tree : Bio.Phylo.BaseTree.Tree, optional
        phylogenetic tree
    cell_order_fields : list, optional
        columns of obs on which to sort cells, by default None
    annotation_fields : list, optional
        column of obs to use as an annotation colorbar, by default 'cluster_id'
    fig : matplotlib.figure.Figure, optional
        existing figure to plot into, by default None
    raw : bool, optional
        raw plotting, no integer color map, by default False
    vmin, vmax : float, optional
        for raw=True, vmin and vmax define the data range that the colormap covers, see `matplotlib.pyplot.imshow`
    max_cn : int, optional
        clip cn at max value, by default 13
    show_cell_ids : bool, optional
        show cell ids on heatmap axis, by default False
    show_subsets : bool, optional
        show subset/superset categoricals to allow identification of cell sets

    Returns
    -------
    dict
        Dictionary of plot and data elements

    Examples
    -------

    .. plot::
        :context: close-figs

        import scgenome
        adata = scgenome.datasets.OV2295_HMMCopy_reduced()

        g = scgenome.pl.plot_cell_cn_matrix_fig(
            adata,
            cell_order_fields=['cell_order'],
            annotation_fields=['cluster_id', 'sample_id', 'quality'])

    """

    if fig is None:
        fig = plt.figure()

    if cell_order_fields is None:
        cell_order_fields = []

    if annotation_fields is None:
        annotation_fields = []

    if var_annotation_fields is None:
        var_annotation_fields = []

    if tree is not None:
        if cell_order_fields is not None and len(cell_order_fields) > 0:
            raise ValueError('cannot provide cell_order_fields and tree')

        # Add phylogenetic ordering to anndata obs
        cell_ids = []
        for a in tree.get_terminals():
            cell_ids.append(a.name)

        adata.obs['phylo_order'] = -1
        for idx, _ in adata.obs.iterrows():
            adata.obs.loc[idx, 'phylo_order'] = cell_ids.index(idx)

        cell_order_fields = ['phylo_order']
        num_phylo = 1
        tree_ax_idx = 0
        heatmap_ax_col_idx = 1

    else:
        num_phylo = 0
        heatmap_ax_col_idx = 0

    # Account for number of var annotations above heatmap
    heatmap_ax_row_idx = len(var_annotation_fields) + 1

    # Account for additional annotation fields that will be added after plotting the matrix
    # when we are adding annotation fields to identify cells as per show_subsets
    num_annotations = len(annotation_fields)
    if show_subsets:
        num_annotations += 2

    num_var_annotations = len(var_annotation_fields)

    fig_main, fig_legends = fig.subfigures(nrows=2, ncols=1, height_ratios=[5, 1], squeeze=True)
    fig_legends.patch.set_alpha(0.0)

    width_ratios = [0.5] * num_phylo + [1] + [0.005] + [0.02] * num_annotations
    height_ratios = [0.02] * num_var_annotations + [0.01] + [1]

    axes = fig_main.subplots(
        nrows=len(height_ratios), ncols=len(width_ratios),
        width_ratios=width_ratios, height_ratios=height_ratios,
        squeeze=False, gridspec_kw=dict(hspace=0.02, wspace=0.02))

    # Turn off axes for all annotation rows and columns
    for ax in axes[:heatmap_ax_row_idx, :].flatten():
        ax.set_axis_off()
    for ax in axes[:, heatmap_ax_col_idx+1:].flatten():
        ax.set_axis_off()

    # Re-enable axes and remove ticks for row annotations
    for ax in axes[heatmap_ax_row_idx, heatmap_ax_col_idx+2:].flatten():
        ax.set_axis_on()
        ax.set_yticks([])

    # Re-enable axes and remove ticks for column annotations
    for ax in axes[:heatmap_ax_row_idx-1, heatmap_ax_col_idx].flatten():
        ax.set_axis_on()
        ax.set_xticks([])

    axes_legends = fig_legends.subplots(
        nrows=1, ncols=1+num_annotations+num_var_annotations, squeeze=False)[0]
    for ax in axes_legends:
        ax.set_axis_off()
        ax.set_alpha(0.0)
        ax.patch.set_alpha(0.0)

    if tree is not None:
        # Plot phylogenetic tree
        tree_ax = axes[heatmap_ax_row_idx, tree_ax_idx]
        tree_ax.spines['top'].set_visible(False)
        tree_ax.spines['right'].set_visible(False)
        tree_ax.spines['bottom'].set_visible(True)
        tree_ax.spines['left'].set_visible(False)
        with plt.rc_context({'lines.linewidth': 0.5}):
            Bio.Phylo.draw(tree, label_func=lambda a: '', axes=tree_ax, do_show=False)
        tree_ax.tick_params(axis='x', labelsize=6)
        tree_ax.set_xlabel('branch length', fontsize=8)
        tree_ax.set_ylabel('')
        tree_ax.set_yticks([])

    heatmap_ax = axes[heatmap_ax_row_idx, heatmap_ax_col_idx]
    ax_legend = axes_legends[0]
    g = plot_cell_cn_matrix(
        adata, layer_name=layer_name,
        cell_order_fields=cell_order_fields,
        ax=heatmap_ax, raw=raw, vmin=vmin, vmax=vmax,
        max_cn=max_cn, show_cell_ids=show_cell_ids)

    adata = g['adata']
    im = g['im']

    if not raw:
        legend_info = {'ax_legend': ax_legend}
        legend_info['legend'] = cn_colors.cn_legend(ax_legend, title=layer_name)

    else:
        legend_info = _plot_continuous_legend(ax_legend, im, layer_name)

    if show_subsets:
        # Need to copy the adata to avoid modifying a view
        adata = adata.copy()
        adata.obs['subset'] = pd.Series(np.mod(np.floor_divide(range(adata.shape[0]), 40), 5), index=adata.obs.index, dtype='category')
        adata.obs['superset'] = pd.Series(np.floor_divide(range(adata.shape[0]), 200), index=adata.obs.index, dtype='category')
        annotation_fields.append('superset')
        annotation_fields.append('subset')

    annotation_info = {}

    for ax, ax_legend, annotation_field in zip(axes[heatmap_ax_row_idx, heatmap_ax_col_idx+2:], axes_legends[1:], annotation_fields):
        if adata.obs[annotation_field].dtype.name in ('category', 'object', 'bool'):
            values = adata.obs[[annotation_field]].values
            annotation_info[annotation_field] = _plot_categorical_annotation(values, ax, ax_legend, annotation_field)

        else:
            values = adata.obs[[annotation_field]].values
            annotation_info[annotation_field] = _plot_continuous_annotation(values, ax, ax_legend, annotation_field)

    for ax, ax_legend, annotation_field in zip(axes[:, heatmap_ax_col_idx], axes_legends[1+len(annotation_fields):], var_annotation_fields):
        if adata.var[annotation_field].dtype.name in ('category', 'object', 'bool'):
            values = adata.var[[annotation_field]].copy().values.T
            annotation_info[annotation_field] = _plot_categorical_annotation(values, ax, ax_legend, annotation_field, horizontal=True)

        else:
            values = adata.var[[annotation_field]].copy().values.T
            annotation_info[annotation_field] = _plot_continuous_annotation(values, ax, ax_legend, annotation_field, horizontal=True)

    return {
        'fig': fig,
        'axes': axes,
        'adata': adata,
        'im': im,
        'legend_info': legend_info,
        'annotation_info': annotation_info,
    }



import numpy as np
import pandas as pd
import pkg_resources


def read_chromosome_lengths(genome_fasta_index):
    chromosome_lengths = pd.read_csv(
        genome_fasta_index, sep='\t',
        header=None, names=['chr', 'chromosome_length', 'V3', 'V4', 'V5'])
    chromosome_lengths = chromosome_lengths[['chr', 'chromosome_length']]
    return chromosome_lengths


class RefGenomeInfo(object):
    def __init__(self, version):
        if version == 'hg19':
            self.chromosomes = [str(a) for a in range(1, 23)] + ['X', 'Y']
            self.plot_chromosomes = [str(a) for a in range(1, 23)] + ['X', 'Y']
            self.genome_fasta_index = pkg_resources.resource_filename('scgenome', 'data/hg19.fa.fai')

        elif version == 'grch38':
            self.chromosomes = [f'chr{a}' for a in range(1, 23)] + ['chrX', 'chrY']
            self.plot_chromosomes = [str(a) for a in range(1, 23)] + ['X', 'Y']
            self.genome_fasta_index = pkg_resources.resource_filename('scgenome', 'data/grch38.fa.fai')

        elif version == 'mm10':
            self.chromosomes = [str(a) for a in range(1, 20)] + ['X', 'Y']
            self.plot_chromosomes = [str(a) for a in range(1, 20)] + ['X', 'Y']
            self.genome_fasta_index = pkg_resources.resource_filename('scgenome', 'data/mm10.fa.fai')

        else:
            raise ValueError()

        self.chromosome_info = read_chromosome_lengths(self.genome_fasta_index)

        # Subset and order according to list of chromosomes
        self.chromosome_info = self.chromosome_info.set_index('chr').loc[self.chromosomes].reset_index()
        self.chromosome_info['chr_index'] = range(self.chromosome_info.shape[0])

        # Add plotting names of chromosomes
        self.chromosome_info['chr_plot'] = self.plot_chromosomes

        # Add start end and mid
        self.chromosome_info['chromosome_end'] = np.cumsum(self.chromosome_info['chromosome_length'])
        self.chromosome_info['chromosome_start'] = self.chromosome_info['chromosome_end'].shift(1)
        self.chromosome_info.loc[self.chromosome_info.index[0], 'chromosome_start'] = 0
        self.chromosome_info['chromosome_start'] = self.chromosome_info['chromosome_start'].astype(int)
        self.chromosome_info['chromosome_mid'] = (self.chromosome_info['chromosome_start'] + self.chromosome_info['chromosome_end']) // 2


info = None


def set_genome_version(version):
    global info
    info = RefGenomeInfo(version)

set_genome_version('hg19')




import sklearn.decomposition
import pandas as pd
import anndata as ad
import numpy as np

import scgenome.preprocessing.transform

from anndata import AnnData


def pca_loadings(adata: AnnData, layer=None, n_components=None, random_state=100) -> AnnData:
    """ Compute PCA loadings matrix

    Parameters
    ----------
    adata : AnnData
        Copy number or other cell matrix
    layer : str, optional
        layer to use, by default None, use .X
    n_components : int, optional
        sklearn.decomposition.PCA n_components parameter, by default None
    random_state : int, optional
        sklearn.decomposition.PCA random_state parameter, by default 100

    Returns
    -------
    DataFrame
        dataframe of pca loadings (rows) by cell (columns)
    """    
    
    """ Calculate loadings of PCA decomposition of a feature matrix.

    Args:
        adata (anndata.AnnData): feature matrix
    """

    if layer is None:
        data = adata.X
    else:
        data = adata.layers[layer]

    pca = sklearn.decomposition.PCA(n_components=n_components, random_state=random_state)
    transformed = pca.fit_transform(scgenome.preprocessing.transform.fill_missing(data))

    var = adata.var.copy()
    var['pca_mean'] = pca.mean_

    obs = pd.DataFrame({
        'component': 'PC' + pd.Series(np.arange(pca.explained_variance_.shape[0]) + 1).astype(str),
        'pca_explained_variance': pca.explained_variance_,
        'pca_explained_variance_ratio': pca.explained_variance_ratio_,
        'pca_singular_values': pca.singular_values_,
    }).set_index('component')

    transformed = pd.DataFrame(
        transformed,
        index=adata.obs.index,
        columns=obs.index)

    uns = {
        'pca': {
            'params': {
                'layer': layer,
                'n_components': n_components,
                'random_state': random_state,
            },
            'results': {
                'n_components': pca.n_components_,
                'n_features': pca.n_features_,
                'n_samples': pca.n_samples_,
                'noise_variance': pca.noise_variance_,
                'n_features_in': pca.n_features_in_,
                'transformed': transformed,
            }
        }
    }

    adata = ad.AnnData(
        pca.components_,
        obs=obs,
        var=var,
        uns=uns,
    )

    return adata



import scipy
import numpy as np
import pandas as pd


def filter_cells(cn, scores, threshold=0.75):
    cn.set_index(['chr', 'start', 'end', 'width'], inplace=True)

    scores = scores[
        (scores['joint'] >= threshold) &
        (scores['cell_call'].isin(['C1', 'C2']))
    ]

    for rm_cond in ['gDNA', 'GM', 'NCC', 'NTC']:
        mask = ~scores['experimental_condition'].str.contains(rm_cond)
        scores = scores[mask]

    good_cells = scores.loc[
        scores['cell_id'].isin(cn.columns.tolist()), 'cell_id'
    ].tolist()
    good_cells.sort()

    return cn[good_cells].reset_index()


def filter_qdnaseq_bins(cn, blacklist):
    blacklist.rename(columns={'chromosome': 'chr'}, inplace=True)

    cn = cn.merge(blacklist, how='left')

    # residual cutoff is 4 * madDiff(residual)
    cn = cn[
        (
            (
                ~pd.isnull(cn['residual']) &
                (cn['residual'].abs() <= 0.0696924)
            ) |
            pd.isnull(cn['residual'])
        ) &
        (cn['blacklist'] == 0)
    ]

    rm_cols = ['bases', 'gc', 'mappability', 'blacklist', 'residual', 'use']
    for rm_col in rm_cols:
        del cn[rm_col]

    return cn


def remove_contiguous_duplicate_bins(cn):
    cn.sort_values(by=['chr', 'start', 'end'], inplace=True)
    cn.set_index(['start', 'end', 'width'], inplace=True)
    cn = cn[(cn.shift() != cn).any(axis=1)]
    cn = cn.reset_index().set_index(['chr', 'start', 'end', 'width'])
    cn.reset_index(inplace=True)
    return cn


def calc_prop_hom_del(states):
    cndist = states.value_counts()
    cndist = cndist / cndist.sum()
    if 0 not in cndist:
        return 0
    return cndist[0]


def calculate_filter_metrics(
        metrics_data,
        cn_data,
        quality_score_threshold=0.75,
        read_count_threshold=500000,
        prop_hom_del_pval_threshold=0.01,
        copy_state_diff_threshold=1.,
    ):
    """ Calculate additional filtering values and add to metrics.
    """
    metrics_data['filter_quality'] = (metrics_data['quality'] > quality_score_threshold)
    metrics_data['filter_reads'] = (metrics_data['total_mapped_reads_hmmcopy'] > read_count_threshold)

    # Calculate proportion homozygous deletion state
    prop_hom_del = cn_data.groupby('cell_id')['state'].apply(calc_prop_hom_del).rename('prop_hom_del').reset_index()
    metrics_data = metrics_data.merge(prop_hom_del, how='left')
    metrics_data['prop_hom_del'] = metrics_data['prop_hom_del'].fillna(0)

    # Calculate p value for each proportion assuming beta fit
    a, b, loc, scale = scipy.stats.beta.fit(metrics_data['prop_hom_del'].values)
    metrics_data['prop_hom_del_pval'] = 1.0 - scipy.stats.beta.cdf(
        metrics_data['prop_hom_del'].values, a, b, loc, scale)
    metrics_data['filter_prop_hom_del'] = (metrics_data['prop_hom_del_pval'] > prop_hom_del_pval_threshold)

    # Calculate separation between predicted and normalized copy number
    copy_state_diff = cn_data[['cell_id', 'copy', 'state']].copy()
    copy_state_diff['copy_state_diff'] = np.absolute(copy_state_diff['copy'] - copy_state_diff['state'])
    copy_state_diff = (copy_state_diff[['cell_id', 'copy_state_diff']]
        .dropna().groupby('cell_id')['copy_state_diff']
        .mean().reset_index().dropna())
    metrics_data = metrics_data.merge(copy_state_diff)
    metrics_data['filter_copy_state_diff'] = (metrics_data['copy_state_diff'] < copy_state_diff_threshold)

    # Filter s phase column
    metrics_data['filter_is_s_phase'] = ~(metrics_data['is_s_phase'].fillna(False))

    return metrics_data



import pyranges as pr
import pandas as pd
import anndata as ad

from pyranges import PyRanges
from pandas import DataFrame
from anndata import AnnData
from numpy import ndarray

import scgenome.refgenome



def dataframe_to_pyranges(data: DataFrame) -> PyRanges:
    """ Convert dataframe with chr, start, end to pyranges object

    Parameters
    ----------
    data : DataFrame
        dataframe with columns 'chr', 'start', 'end'

    Returns
    -------
    PyRanges
        pyranges object
    """
    data = pr.PyRanges(data.rename(columns={
        'chr': 'Chromosome',
        'start': 'Start',
        'end': 'End',
    }))

    return data


def pyranges_to_dataframe(data: PyRanges) -> DataFrame:
    """ Convert pyranges object to dataframe

    Parameters
    ----------
    data : PyRanges
        pyranges object

    Returns
    -------
    DataFrame
        dataframe with columns 'chr', 'start', 'end'
    """
    data = data.as_df().rename(columns={
        'Chromosome': 'chr',
        'Start': 'start',
        'End': 'end',
    })

    return data


def create_bins_pr(binsize: int) -> PyRanges:
    """ Create a regular binning of the genome

    Parameters
    ----------
    binsize : int
        size of bins

    Returns
    -------
    PyRanges
        regular bins tiled across the genome
    """    
    
    chromsizes = scgenome.refgenome.info.chromosome_info[['chr', 'chromosome_length']].rename(
        columns={'chr': 'Chromosome', 'chromosome_length': 'End'}).assign(Start=0)[['Chromosome', 'Start', 'End']]

    chromsizes = pr.PyRanges(chromsizes)

    bins = pr.gf.tile_genome(chromsizes, binsize)

    return bins


def create_bins(binsize: int) -> DataFrame:
    """ Create a regular binning of the genome

    Parameters
    ----------
    binsize : int
        size of bins

    Returns
    -------
    DataFrame
        regular bins tiled across the genome
    """
    return pyranges_to_dataframe(create_bins_pr(binsize=binsize))


def rebin_agg_df(data: DataFrame, intersect: DataFrame, agg_f: dict) -> DataFrame:
    """ Rebin a dataframe, aggregating across multiple intersecting regions

    Parameters
    ----------
    data : DataFrame
        binned data to rebin and aggregate with columns 'chr', 'start', 'end' and index 'bin'
    intersect : DataFrame
        mapping between previous and target bins with columns 'bin', 'target_bin', 'width'
    agg_f : dict of tuples
        aggregate functions, similar to pandas dataframe groupby .agg

    Returns
    -------
    DataFrame
        rebinned and aggregated data with columns according to 'agg_f' and index 'target_bin'
    """
    data = data.loc[intersect['bin'], :]

    data = data.set_index(intersect.set_index('bin')['target_bin'], append=True)
    data = data.set_index(intersect.set_index('bin')['width'], append=True)
    
    data = data.groupby(level='target_bin').agg(**agg_f)

    return data


def rebin_agg_layer(adata: AnnData, intersect: DataFrame, layer_name: str, agg_f: dict) -> DataFrame:
    """ Rebin an anndata layer, aggregating across multiple intersecting regions

    Parameters
    ----------
    adata : DataFrame
        adata to rebin and aggregate with var columns 'chr', 'start', 'end' and index 'bin'
    intersect : DataFrame
        mapping between previous and target bins with columns 'bin', 'target_bin', 'width'
    layer_name : str
        name of layer to rebin and aggregate
    agg_f : dict of tuples
        aggregate functions, similar to pandas dataframe groupby .agg

    Returns
    -------
    DataFrame
        rebinned and aggregated layer data with columns according to 'agg_f' and index 'target_bin'
    """
    data = adata.to_df(layer=layer_name)

    data = data.loc[:, intersect['bin']].T

    data = data.set_index(intersect.set_index('bin')['target_bin'], append=True)
    data = data.set_index(intersect.set_index('bin')['width'], append=True)

    data = data.groupby(level='target_bin').agg(agg_f)
    
    return data.T


def intersect_regions(a: DataFrame, b: DataFrame) -> DataFrame:
    """ Compute intersection between two sets of regions

    Parameters
    ----------
    a : DataFrame
        region data with columns 'chr', 'start', 'end', ...
    b : DataFrame
        region data with columns 'chr', 'start', 'end', ...

    Returns
    -------
    DataFrame
        intersecting regions with columns 'chr', 'start', 'end' and columns from 'a' and 'b'
    """
    a = scgenome.tools.ranges.dataframe_to_pyranges(a)
    b = scgenome.tools.ranges.dataframe_to_pyranges(b)

    intersect_1 = a.intersect(b)
    intersect_2 = b.intersect(a)

    intersect = pd.merge(
        scgenome.tools.ranges.pyranges_to_dataframe(intersect_1),
        scgenome.tools.ranges.pyranges_to_dataframe(intersect_2),
        on=['chr', 'start', 'end'])

    return intersect


def rebin(adata: AnnData, target_bins: DataFrame, outer_join: bool=False, agg_X=None, agg_var=None, agg_layers=()) -> AnnData:
    """ Rebin an AnnData and aggregate across multiple intersecting regions

    Parameters
    ----------
    adata : AnnData
        data to rebin and aggregate
    target_bins : DataFrame
        target bins with columns 'chr', 'start', 'end'
    outer_join : bool, optional
        whether to include target bins with no overlap in adata, by default False
    agg_X : callable, optional
        aggregate function for X, by default None
    agg_var : dict, optional
        aggregate functions for var, by default None
    agg_layers : dict, optional
        aggregate functions for each layer, by default ()

    Returns
    -------
    AnnData
        rebinned and aggregated data
    """
    bins = adata.var.rename_axis('bin').reset_index()[['chr', 'start', 'end', 'bin']]

    target_bins['target_bin'] = (
        target_bins['chr'].astype(str) + ':' +
        target_bins['start'].astype(str) + '-' +
        target_bins['end'].astype(str))
    target_bins = target_bins[['chr', 'start', 'end', 'target_bin']]

    intersect = intersect_regions(bins, target_bins)
    intersect['width'] = intersect['end'] - intersect['start'] + 1

    if agg_var is not None:
        var = rebin_agg_df(adata.var, intersect, agg_var)
    
    else:
        var = intersect[['target_bin']].set_index('target_bin')

    var = target_bins.set_index('target_bin').merge(
        var, left_index=True, right_index=True, how=('inner', 'left')[outer_join])

    var = var.sort_values(['chr', 'start'])

    if agg_X is not None:
        X = rebin_agg_layer(adata, intersect, None, agg_X)
        X = X.reindex(columns=var.index)

    else:
        X = None

    layer_data = {}
    for layer_name in agg_layers:
        layer_data[layer_name] = rebin_agg_layer(adata, intersect, layer_name, agg_layers[layer_name])
        layer_data[layer_name] = layer_data[layer_name].reindex(columns=var.index)

    adata = ad.AnnData(
        X,
        obs=adata.obs,
        var=var,
        layers=layer_data,
    )

    return adata


def rebin_regular(adata: AnnData, bin_size: int, outer_join: bool=False, agg_X=None, agg_var=None, agg_layers=()) -> AnnData:
    """ Rebin an AnnData and aggregate across multiple intersecting regions

    Parameters
    ----------
    adata : AnnData
        data to rebin and aggregate
    bin_size : int
        width of target bins
    outer_join : bool, optional
        whether to include target bins with no overlap in adata, by default False
    agg_X : callable, optional
        aggregate function for X, by default None
    agg_var : dict, optional
        aggregate functions for var, by default None
    agg_layers : dict, optional
        aggregate functions for each layer, by default ()

    Returns
    -------
    AnnData
        rebinned and aggregated data
    """
    target_bins = create_bins(bin_size)
    target_bins['start'] = target_bins['start'] + 1

    return rebin(adata, target_bins, outer_join=outer_join, agg_X=agg_X, agg_var=agg_var, agg_layers=agg_layers)


def weighted_mean(values: ndarray, widths: ndarray) -> float:
    """ Compute weighted mean

    Parameters
    ----------
    values : ndarray
        values to compute mean 
    widths : ndarray
        weights corresponding to each value

    Returns
    -------
    float
        weighted mean
    """
    weighted_mean = (widths * values).sum() / widths.sum()
    return weighted_mean


def bin_width_weighted_mean(df: DataFrame) -> float:
    """ Convenience function to allow computing of weighted mean with rebin aggregate functions

    Parameters
    ----------
    df : DataFrame
        group of data for which to compute mean, must have width in index

    Returns
    -------
    float
        weighted mean
    """
    values = df.values
    widths = df.index.get_level_values('width').values
    return weighted_mean(values, widths)





import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn
import wgs_analysis.plots.rearrangement
import wgs_analysis.algorithms.merge

import scgenome.cntransitions


def get_index_cols(is_lumpy=False):
    if is_lumpy:
        return ['breakpoint_id']
    return ['prediction_id']


def get_bp_index_cols(is_lumpy=False):
    if is_lumpy:
        return ['chrom1', 'strand1', 'start1',
                'chrom2', 'strand2', 'start2'
                ]
    return ['chromosome_1', 'strand_1', 'position_1',
            'chromosome_2', 'strand_2', 'position_2'
            ]


def annotate_breakpoint_data(breakpoint_data, breakpoint_count_data, is_lumpy=False):
    """ Load breakpoints and add annotations.
    """

    index_cols = get_index_cols(is_lumpy=is_lumpy)

    evidence_count_colname = "read_count"
    if is_lumpy:
        evidence_count_colname = "count"

    # Calculate cell counts
    if len(breakpoint_count_data.index) > 0:
        cell_counts = (
            breakpoint_count_data
            .query('{} > 0'.format(evidence_count_colname))
            .drop_duplicates(index_cols + ['cell_id'])
            .groupby(index_cols).size().rename('num_cells')
            .reset_index())

        breakpoint_data = breakpoint_data.merge(cell_counts, how='left')
        
        assert not breakpoint_data['num_cells'].isnull().any()

    else:
        breakpoint_data['num_cells'] = 0

    return breakpoint_data, breakpoint_count_data


def filter_breakpoint_data(
        breakpoint_data,
        breakpoint_count_data,
        num_split_threshold=1,
        template_length_min_threshold=250,
        num_cells_threshold=2,
        num_unique_reads_threshold=2,
        lumpy=False,
    ):
    """ Filter breakpoint data and breakpoint counts
    """
    index_cols = get_index_cols(is_lumpy=lumpy)
    bp_index_cols = get_bp_index_cols(is_lumpy=lumpy)

    num_initial_breakpoints = len(breakpoint_data.index)

    filtered_predictions = []

    # Filter by number of split reads
    if num_split_threshold is not None:
        filtered_num_split = breakpoint_data.query(
            'num_split < {}'.format(num_split_threshold))
        logging.info('Filtering {} of {} breakpoints by num_split < {}'.format(
            len(filtered_num_split.index), num_initial_breakpoints, num_split_threshold))
        filtered_predictions.append(filtered_num_split[index_cols].drop_duplicates())

    # Filter by predicted sequence length
    if template_length_min_threshold is not None:
        filtered_template_length_min = breakpoint_data.query(
            'template_length_min < {}'.format(template_length_min_threshold))
        logging.info('Filtering {} of {} breakpoints by template_length_min < {}'.format(
            len(filtered_template_length_min.index), num_initial_breakpoints, template_length_min_threshold))
        filtered_predictions.append(filtered_template_length_min[index_cols].drop_duplicates())

    # Filter breakpoints by num cells in which they are detected
    if num_cells_threshold is not None:
        filtered_cell_counts = breakpoint_data.query(
            'num_cells < {}'.format(num_cells_threshold))
        logging.info('Filtering {} of {} breakpoints by num_cells < {}'.format(
            len(filtered_cell_counts.index), num_initial_breakpoints, num_cells_threshold))
        filtered_predictions.append(filtered_cell_counts[index_cols].drop_duplicates())

    # Filter breakpoints by total read counts
    if num_unique_reads_threshold is not None:
        filtered_num_unique_reads = breakpoint_data.query(
            'num_unique_reads < {}'.format(num_unique_reads_threshold))
        logging.info('Filtering {} of {} breakpoints by num_unique_reads < {}'.format(
            len(filtered_num_unique_reads.index), num_initial_breakpoints, num_unique_reads_threshold))
        filtered_predictions.append(filtered_num_unique_reads[index_cols].drop_duplicates())

    if len(filtered_predictions) > 0:
        filtered_predictions = pd.concat(filtered_predictions, ignore_index=True).drop_duplicates()
        logging.info('Filtering {} of {} breakpoints'.format(
            len(filtered_predictions.index), num_initial_breakpoints))
        breakpoint_data = breakpoint_data[
            ~breakpoint_data.set_index(index_cols).index.isin(filtered_predictions.set_index(index_cols).index)]
        logging.info('Retained {} of {} breakpoints'.format(
            len(breakpoint_data.index), num_initial_breakpoints))

    else:
        logging.info('No filtering applied')

    breakpoint_count_data = breakpoint_count_data.merge(
        breakpoint_data[index_cols + bp_index_cols].drop_duplicates())

    return breakpoint_data, breakpoint_count_data


def plot_library_portrait(breakpoint_data, figures_prefix=None):
    """ Plot a library level portrait of breakpoint features.
    """

    breakpoint_data['log_num_reads'] = np.log10(breakpoint_data['num_reads'].astype(float))
    breakpoint_data['log_num_split'] = np.log10(breakpoint_data['num_split'].astype(float))
    breakpoint_data['log_num_cells'] = np.log10(breakpoint_data['num_cells'].astype(float))

    # Plot per library feature distributions
    features = [
        ('log_num_reads', 'log discordant read counts'),
        ('log_num_split', 'log split read counts'),
        ('template_length_min', 'prediction sequence length'),
        ('homology', 'sequence homology'),
        ('log_num_cells', 'log number of cells'),
    ]

    order = breakpoint_data['library_id'].unique()
    hue_order = breakpoint_data['rearrangement_type'].unique()

    fig = plt.figure(figsize=(8, 12))
    ax = fig.add_subplot(len(features) + 1, 1, 1)
    plot_data = (
        breakpoint_data
        .groupby(['library_id', 'rearrangement_type'])
        .size().rename('count').reset_index())
    seaborn.barplot(
        ax=ax, data=plot_data,
        x='library_id', y='count', hue='rearrangement_type',
        order=order, hue_order=hue_order)
    ax.set_title(f'Counts by rearrangement type across libraries')
    ax.set_xticklabels([])
    ax.set_xlabel('', visible=False)
    ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

    for idx, (col, desc) in enumerate(features):
        ax = fig.add_subplot(len(features) + 1, 1, idx + 2)
        plot_data = breakpoint_data[['library_id', 'rearrangement_type', col]].dropna()
        plot_data[col] = plot_data[col].astype(float)
        seaborn.boxplot(
            ax=ax, data=plot_data,
            x='library_id', y=col, hue='rearrangement_type',
            order=order, hue_order=hue_order)
        ax.set_title(f'Distribution of {desc} across libraries')
        if idx < len(features) - 1:
            ax.set_xticklabels([])
            ax.set_xlabel('', visible=False)
        ax.legend().set_visible(False)

    plt.tight_layout()
    if figures_prefix is not None:
        fig.savefig(figures_prefix + 'library_features.pdf', bbox_inches='tight')

    # Plot rearrangement type distribution across the genome per library
    num_libraries = len(breakpoint_data['library_id'].unique())
    fig = plt.figure(figsize=(10, 3 * num_libraries))
    for idx, (library_id, library_breakpoints) in enumerate(breakpoint_data.groupby('library_id', observed=True)):
        breakends = wgs_analysis.algorithms.rearrangement.create_breakends(
            library_breakpoints, data_cols=['rearrangement_type'])
        ax = fig.add_subplot(num_libraries, 1, idx + 1)
        wgs_analysis.plots.rearrangement.chromosome_type_plot(
            ax, breakends)
        ax.set_title(f'Chromosome types for library {library_id}')
    plt.tight_layout()
    if figures_prefix is not None:
        fig.savefig(figures_prefix + 'chromosome_types.pdf', bbox_inches='tight')

    # Plot adjacent density of breakends across the genome per library
    fig = plt.figure(figsize=(10, 3 * num_libraries))
    for idx, (library_id, library_breakpoints) in enumerate(breakpoint_data.groupby('library_id')):
        breakends = wgs_analysis.algorithms.rearrangement.create_breakends(
            library_breakpoints)
        breakends['chrom'] = breakends['chromosome']
        breakends['coord'] = breakends['position']
        breakends = wgs_analysis.annotation.position.annotate_adjacent_distance(breakends)
        ax = fig.add_subplot(num_libraries, 1, idx + 1)
        wgs_analysis.plots.snv.snv_adjacent_distance_plot(
            ax, breakends)
        ax.set_title(f'Breakend adjacent distances for library {library_id}')
    plt.tight_layout()
    if figures_prefix is not None:
        fig.savefig(figures_prefix + 'adjacent_distance.pdf', bbox_inches='tight')


def plot_breakpoint_clustering(breakpoint_data, breakpoint_count_data, clusters, figures_prefix=None):
    """ Plot breakpoint cluster figures.
    """
    bp_index_cols = get_bp_index_cols(is_lumpy=False)

    plot_data = breakpoint_count_data.merge(clusters)
    plot_data = (
        plot_data.groupby(bp_index_cols + ['cluster_id'])[['read_count']]
        .sum().unstack(fill_value=None).stack().reset_index())

    g = seaborn.factorplot(y='read_count', x='cluster_id', kind='box', data=plot_data, color='0.75', size=4)
    if figures_prefix is not None:
        g.fig.savefig(figures_prefix + 'cluster_id_read_counts.pdf', bbox_inches='tight')

    plot_data['is_present'] = (plot_data['read_count'] > 0) * 1
    plot_data = plot_data.set_index(bp_index_cols + ['cluster_id'])['is_present'].unstack(fill_value=None)
    mask = plot_data.isnull()
    plot_data = plot_data.fillna(0)

    g = seaborn.clustermap(plot_data, mask=mask, rasterized=True, figsize=(4, 12))
    if figures_prefix is not None:
        g.fig.savefig(figures_prefix + 'cluster_map.pdf', bbox_inches='tight')


def overlap_regions_positions_stranded(regions, positions):
    """ Find overlaps between regions and positions accounting for strand/orientation.

    Args:
        regions (DataFrame): oriented genomic regions
        positions (DataFrame): genomic positions with orientation as strand

    Returns:
        DataFrame: cross product of overlapping regions and positions

    """

    merged = []
    for (chromosome, strand), chr_str_posns in positions.groupby(['chromosome', 'strand']):
        chr_str_regions = regions[
            (regions['chr'] == chromosome) &
            (regions['orientation'] == strand)]
        interval_idx, position_idx = wgs_analysis.algorithms.merge.interval_position_overlap_unsorted(
            chr_str_regions[['start', 'end']].values,
            chr_str_posns['position'].values)
        chr_str_merged = pd.concat([
            chr_str_regions.iloc[interval_idx].reset_index(drop=True),
            chr_str_posns.iloc[position_idx].reset_index(drop=True)], axis=1)
        merged.append(chr_str_merged)
    merged = pd.concat(merged, ignore_index=True)
    return merged


def overlap_regions_breakpoints_stranded(regions, breakpoints, cols):
    """ Find overlap between breakends and regions.

    Args:
        regions (DataFrame): oriented genomic regions
        breakpoints (DataFrame): breakpoints table
        cols (list): list of columns to include in annotated breakpoint dataframe

    Returns:
        DataFrame: breakpoints with additional columns from overlapping with regions

    """

    for end in ('1', '2'):
        breakends = breakpoints[['prediction_id', f'chromosome_{end}', f'strand_{end}', f'position_{end}']].rename(
            columns={
                f'chromosome_{end}': 'chromosome',
                f'strand_{end}': 'strand',
                f'position_{end}': 'position'})[['prediction_id', 'chromosome', 'strand', 'position']]
        merged = overlap_regions_positions_stranded(regions, breakends)
        end_cols = [f'{a}_{end}' for a in cols]
        merged = merged.rename(columns=dict(zip(cols, end_cols)))
        breakpoints = breakpoints.merge(merged[['prediction_id'] + end_cols], how='left', on=['prediction_id'])
    return breakpoints


def annotate_transition_cell_counts(breakpoints, cn_data):
    """ Annotate breakpoints with counts of cells supporting matching transitions.

    Args:
        breakpoints (DataFrame): breakpoints table
        cn_data (DataFrame): copy number data
    
    Returns:
        DataFrame: breakpoints table with counts of supporting cells

    """

    cn_transition_counts = scgenome.cntransitions.generate_cn_transition_counts(cn_data)

    breakpoints = scgenome.breakpointdata.overlap_regions_breakpoints_stranded(
        cn_transition_counts, breakpoints,
        ['region_index', 'cell_count'])

    breakpoints['cell_count_1'] = breakpoints['cell_count_1'].fillna(0).astype(int)
    breakpoints['cell_count_2'] = breakpoints['cell_count_2'].fillna(0).astype(int)

    cn_transitions_matrix = scgenome.cntransitions.generate_cn_transition_pair_counts(cn_data)

    breakpoints = breakpoints.merge(cn_transitions_matrix, how='left')
    breakpoints['pair_cell_count'] = breakpoints['pair_cell_count'].fillna(0).astype(int)

    breakpoints = breakpoints.drop(['region_index_1', 'region_index_2'], axis=1)

    return breakpoints



""" Single-Cell Whole Genome Analysis in Python. """

from . import tools as tl
from . import preprocessing as pp
from . import plotting as pl
from . import datasets

from . import _version
__version__ = _version.get_versions()['version']

# has to be done at the end, after everything has been imported
import sys

sys.modules.update({f'{__name__}.{m}': globals()[m] for m in ['tl', 'pp', 'pl']})

del sys



import os
from collections import defaultdict

import pandas as pd
import scgenome.loaders.utils
import scgenome.utils
import yaml
from csverve.core import CsverveInput


_categorical_cols = [
    'cell_id',
    'sample_id',
    'library_id',
]


def load_annotation_files(annotation_metrics):
    results_tables = {}

    results_tables['annotation_metrics'] = process_annotation_file(annotation_metrics)

    scgenome.utils.union_categories(results_tables.values())

    return results_tables


def load_annotation_results(
        results_dir,
):
    """ Load annotation tables
    
    Args:
        results_dir (str): results directory to load from.
    
    Returns:
        dict: pandas.DataFrame tables keyed by table name
    """

    annotation_metrics_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, '_metrics.csv.gz', analysis_type='annotation')

    return load_annotation_files(annotation_metrics_filepath)


def process_annotation_file(filepath):
    data = CsverveInput(filepath).read_csv()

    data['sample_id'] = [a.split('-')[-4] for a in data['cell_id']]
    data['library_id'] = [a.split('-')[-3] for a in data['cell_id']]

    for col in _categorical_cols:
        if col in data:
            data[col] = pd.Categorical(data[col])

    return data



import numpy as np
from remixt.bpmodel import sum_product
from remixt.bpmodel import sum_product_2paramtrans
import scipy


def calculate_ll_normal_simple(data, variances):
    """ Calculate likelihood per state per segment
    
    Args:
        data: (n_cells, n_segments)
        variances: (n_cells, n_states)

    Returns:
        log likelihood: (n_cells, n_segments, n_states)
    """

    n_cells = data.shape[0]
    n_segments = data.shape[1]
    n_states = variances.shape[1]

    # Create states as (n_segments, n_cells, n_states) array
    states = np.tile(np.arange(0, n_states, 1), (n_cells, n_segments, 1))

    # Calculate mean
    mean = states
    
    # Normal dist log likelihood
    ll = (
        -0.5 * np.log(2. * np.pi)
        -0.5 * np.log(variances[:, np.newaxis, :])
        -1. * (np.square(data[:, :, np.newaxis] - mean) /
                (2. * variances[:, np.newaxis, :])))
    ll[np.isnan(data)] = 0.

    return ll


def calculate_marginal_ll_simple(data, variances, transmodel):
    framelogprob = calculate_ll_normal_simple(data, variances).sum(axis=0)

    alphas = np.zeros(framelogprob.shape)
    betas = np.zeros(framelogprob.shape)

    if transmodel['kind'] == 'twoparam':
        sum_product_2paramtrans(
            framelogprob,
            alphas,
            betas,
            transmodel['e0'],
            transmodel['e1'],
        )

    elif transmodel['kind'] == 'full':
        sum_product(
            framelogprob,
            transmodel['tr_mat'],
            alphas,
            betas,
        )

    else:
        raise ValueError("unknown transition model {transmodel['kind']}")

    return scipy.special.logsumexp(alphas[-1, :])


def gibbs_sample_cluster_indices(data, variances, assignments, max_clusters, alpha, transmodel):
    n_cells = data.shape[0]

    for cell_idx in range(n_cells):
        log_assign_prob = np.zeros((max_clusters,))

        for cluster_idx in range(max_clusters):
            assignments[cell_idx] = cluster_idx
            log_marginal_with = calculate_marginal_ll_simple(
                data[assignments == cluster_idx, :],
                  variances[assignments == cluster_idx, :],
                transmodel)

            assignments[cell_idx] = -1
            log_marginal_without = calculate_marginal_ll_simple(
                data[assignments == cluster_idx, :],
                variances[assignments == cluster_idx, :],
                transmodel)

            log_posterior_predictive = log_marginal_with - log_marginal_without
            num_cluster = np.sum(assignments[cell_idx] == cluster_idx)
            if num_cluster == 0:
                log_assign_prob[cluster_idx] = np.log(alpha / (alpha + n_cells - 1)) + log_posterior_predictive
            else:
                log_assign_prob[cluster_idx] = np.log(num_cluster / (alpha + n_cells - 1)) + log_posterior_predictive
        
        assign_prob = np.exp(log_assign_prob - log_assign_prob.max())
        assign_prob /= assign_prob.sum()

        assignments[cell_idx] = np.random.choice(assign_prob.shape[0], p=assign_prob)
    
    return assignments





import os
from collections import defaultdict

import pandas as pd
import scgenome.loaders.utils
import scgenome.utils
import yaml
from csverve.core import CsverveInput



_categorical_cols = [
    'cell_id',
    'sample_id',
    'library_id',
]


def load_alignment_files(align_metrics, gc_metrics=None):
    results_tables = dict()

    results_tables["align_metrics"] = process_alignment_data(align_metrics)

    if gc_metrics:
        results_tables["gc_metrics"] = process_alignment_data(gc_metrics)

    scgenome.utils.union_categories(results_tables.values())

    return results_tables


def load_alignment_results(results_dir):
    """ Load alignment metrics tables
    
    Args:
        results_dir (str): results directory to load from.
    
    Returns:
        dict: pandas.DataFrame tables keyed by table name
    """

    alignment_metrics_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, '_alignment_metrics.csv.gz', 'alignment_metrics', analysis_type='alignment')

    gc_metrics_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, '_gc_metrics.csv.gz', 'alignment_gc_metrics', analysis_type='alignment')

    return load_alignment_files(alignment_metrics_filepath, gc_metrics=gc_metrics_filepath)


def process_alignment_data(filepath):
    data = CsverveInput(filepath).read_csv()

    data.query(f"cell_id != 'reference'", inplace=True)

    data['sample_id'] = [a.split('-')[-4] for a in data['cell_id']]
    data['library_id'] = [a.split('-')[-3] for a in data['cell_id']]

    for col in _categorical_cols:
        if col in data:
            data[col] = pd.Categorical(data[col])

    return data



import pandas as pd
import numpy as np
import collections

from . import refgenome


def union_categories(dfs, cat_cols=None):
    """ Recreate specified categoricals on the union of categories inplace.

    Args:
        dfs (list of pandas.DataFrame): pandas dataframes to unify categoricals in-place.
    
    KwArgs:
        cat_cols (list of str): columns to unify categoricals, default None for any categorical.
    """

    # Infer all categorical columns if not given
    if cat_cols is None:
        cat_cols = set()
        for df in dfs:
            for col in df:
                if df[col].dtype.name == 'category':
                    cat_cols.add(col)

    # Get a list of categories for each column
    col_categories = collections.defaultdict(set)
    for col in cat_cols:
        for df in dfs:
            if col in df:
                col_categories[col].update(df[col].values)

    # Remove None and nan as they cannot be in the list of categories
    for col in col_categories:
        col_categories[col] = col_categories[col] - set([None, np.nan])

    # Create a pandas index for each set of categories
    for col, categories in col_categories.items():
        col_categories[col] = pd.Index(categories)

    # Set all categorical columns as having teh same set of categories
    for col in cat_cols:
        for df in dfs:
            if col in df:
                df[col] = df[col].astype('category')
                df[col] = df[col].cat.set_categories(col_categories[col])


def concat_with_categories(dfs, **kwargs):
    """ Concatenate dataframes retaining categorical columns
    """
    # Infer all categorical columns
    cat_cols = set()
    for df in dfs:
        for col in df:
            if df[col].dtype.name == 'category':
                cat_cols.add(col)

    # Get a list of categories for each column
    col_categories = collections.defaultdict(set)
    for df in dfs:
        for col in cat_cols:
            col_categories[col].update(df[col].values)

    # Remove None and nan as they cannot be in the list of categories
    for col in col_categories:
        col_categories[col] = col_categories[col] - set([None, np.nan])

    # Create a pandas index for each set of categories
    for col, categories in col_categories.items():
        col_categories[col] = pd.Index(categories)

    # Set all categorical columns as having teh same set of categories
    for df in dfs:
        for col in cat_cols:
            df[col] = df[col].astype('category')
            df[col] = df[col].cat.set_categories(col_categories[col])

    return pd.concat(dfs, **kwargs)





import numpy as np

from numpy import ndarray


def fill_missing(data: ndarray) -> ndarray:
    """ Fill missing values with means of non-nan values across rows.

    Deal with missing values by assigning the mean value
    of each bin to missing values of that bin

    Parameters
    ----------
    data : ndarray
        data to fill

    Returns
    -------
    ndarray
        copy of data with missing entries filled 
    """    

    data = data.copy()

    # Set bins with nan across all cells to 0
    data[:, np.all(np.isnan(data), axis=0)] = 0

    # Mean of each bin, ignoring nan
    bin_means = np.nanmean(data, axis=0)
    bin_means = np.nan_to_num(bin_means, nan=0)
    bin_means = np.tile(bin_means, (data.shape[0], 1))
    data[np.where(np.isnan(data))] = bin_means[np.where(np.isnan(data))]

    return data




import logging
import scipy.stats
import seaborn
import umap
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import scgenome.cncluster
import scgenome.cnplot
import scgenome.cnfilter


def calculate_umap_hdbscan_clones(cn_data, metrics_data, results_prefix=None):
    """ Cluster copy number data.
    """
    logging.info('creating copy number matrix')
    cn = (
        cn_data
            .set_index(['chr', 'start', 'end', 'cell_id'])['copy']
            .unstack(level='cell_id').fillna(0)
    )

    logging.info('clustering copy number')
    clusters = scgenome.cncluster.umap_hdbscan_cluster(cn)

    num_clusters = len(clusters['cluster_id'].unique())

    fig = plt.figure(figsize=(4, 4))
    scgenome.cncluster.plot_umap_clusters(plt.gca(), clusters)
    if results_prefix is not None:
        fig.savefig(results_prefix + 'cn_umap.pdf', bbox_inches='tight')

    fig = plt.figure(figsize=(num_clusters/4, 4))
    seaborn.barplot(x='cluster_id', y='count', data=clusters.groupby('cluster_id').size().rename('count').reset_index())
    if results_prefix is not None:
        fig.savefig(results_prefix + 'clone_size.pdf', bbox_inches='tight')

    plot_data = (clusters
        .merge(metrics_data[['cell_id', 'sample_id']].drop_duplicates())
        .groupby(['cluster_id', 'sample_id']).size().unstack(fill_value=0).T)
    fig = plt.figure(figsize=(2, num_clusters/4))
    seaborn.heatmap(plot_data.T, annot=True, fmt="d", linewidths=.5)
    if results_prefix is not None:
        fig.savefig(results_prefix + 'clone_sample_size.pdf', bbox_inches='tight')

    logging.info('merging clusters')
    cn_data = cn_data.merge(clusters[['cell_id', 'cluster_id']].drop_duplicates())

    plot_clones(cn_data, 'cluster_id', plots_prefix=results_prefix)

    return clusters


def calculate_kmeans_clones(cn_data, metrics_data, results_prefix=None, min_k=2, max_k=100, min_cluster_size=None):
    """ Cluster copy number data.
    """
    logging.info('creating copy number matrix')
    cn = (
        cn_data
            .set_index(['chr', 'start', 'end', 'cell_id'])['copy']
            .unstack(level='cell_id').fillna(0)
    )

    logging.info('clustering copy number')
    clusters = scgenome.cncluster.kmeans_cluster(cn, min_k=min_k, max_k=max_k)
    clusters = clusters.merge(clusters.groupby('cluster_id').size().rename('cluster_size').reset_index())

    if min_cluster_size is not None:
        clusters = clusters.query(f"cluster_size >= {min_cluster_size}")

    embedding = umap.UMAP(
        n_neighbors=15,
        min_dist=0.1,
        n_components=2,
        random_state=42,
        metric='euclidean',
    ).fit_transform(cn.fillna(0).values.T)

    clusters = clusters.merge(pd.DataFrame({
        'cell_id': cn.columns,
        'umap1': embedding[:, 0], 'umap2': embedding[:, 1]
    }))

    num_clusters = len(clusters['cluster_id'].unique())

    fig = plt.figure(figsize=(4, 4))
    scgenome.cncluster.plot_umap_clusters(plt.gca(), clusters)
    if results_prefix is not None:
        fig.savefig(results_prefix + 'cn_umap.pdf', bbox_inches='tight')

    fig = plt.figure(figsize=(num_clusters/4, 4))
    seaborn.barplot(x='cluster_id', y='count', data=clusters.groupby('cluster_id').size().rename('count').reset_index())
    if results_prefix is not None:
        fig.savefig(results_prefix + 'clone_size.pdf', bbox_inches='tight')

    plot_data = (clusters
        .merge(metrics_data[['cell_id', 'sample_id']].drop_duplicates())
        .groupby(['cluster_id', 'sample_id']).size().unstack(fill_value=0).T)
    fig = plt.figure(figsize=(2, num_clusters/4))
    seaborn.heatmap(plot_data.T, annot=True, fmt="d", linewidths=.5)
    if results_prefix is not None:
        fig.savefig(results_prefix + 'clone_sample_size.pdf', bbox_inches='tight')

    logging.info('merging clusters')
    cn_data = cn_data.merge(clusters[['cell_id', 'cluster_id']].drop_duplicates())

    plot_clones(cn_data, 'cluster_id', plots_prefix=results_prefix)

    return clusters


def breakpoint_filter(metrics_data, clusters, max_breakpoints, results_prefix):
    """ Filter clusters based on breakpoint counts
    """
    breakpoint_data = (
        metrics_data
        .merge(clusters[['cell_id', 'cluster_id']])
        .groupby('cluster_id')['breakpoints']
        .mean().reset_index()
        .sort_values('breakpoints'))

    fig = plt.figure(figsize=(4, 4))
    breakpoint_data['breakpoints'].hist(bins=40)
    fig.savefig(results_prefix + 'breakpoint_hist.pdf', bbox_inches='tight')

    breakpoint_data = breakpoint_data[breakpoint_data['breakpoints'] <= max_breakpoints]

    clusters = clusters[clusters['cluster_id'].isin(breakpoint_data['cluster_id'])]

    return clusters


def finalize_clusters(
        cn_data, metrics_data, clusters, filter_metrics,
        cell_clone_distances, results_prefix,
        is_original_cluster_mean_threshold=0.5,
        cluster_size_threshold=50):
    """ Generate finalized filtered clusters
    """

    # Calculate the cluster assignment based on correlation
    correlation_metric = 'pearsonr'
    correlation_cluster = (
        cell_clone_distances
        .set_index(['cluster_id', 'cell_id'])[correlation_metric]
        .unstack().idxmin().rename(correlation_metric + 'cluster_id').reset_index())

    # Calculate which cells cluster assignment matches highest correlation cluster
    cluster_annotation = cell_clone_distances.merge(clusters[['cell_id', 'cluster_id']])
    cluster_annotation = cluster_annotation.merge(correlation_cluster)
    cluster_annotation['is_original'] = (cluster_annotation['cluster_id'] == cluster_annotation[correlation_metric + 'cluster_id'])

    # Plot the cityblock distance distribution of each cluster separated
    # by whether the assigned cluster equals the correlation cluster
    plot_metric = 'cityblock'
    g = seaborn.factorplot(
        x='cluster_id', y=plot_metric,
        hue='is_original', kind='strip',
        dodge=True, data=cluster_annotation, aspect=3)
    g.fig.savefig(results_prefix + 'cluster_cityblock_distance.pdf', bbox_inches='tight')

    # Calculate the proportion of each cluster that would be assigned to
    # that cluster by maximizing correlation
    cluster_annotation['is_original_f'] = cluster_annotation['is_original'] * 1.
    is_original_mean = cluster_annotation.groupby('cluster_id')['is_original_f'].mean().rename('is_original_cluster_mean').reset_index()
    cluster_annotation = cluster_annotation.merge(is_original_mean)

    # Filter cells that are not assigned to the same cluster they
    # are most correlated with
    cluster_annotation = cluster_annotation.query('is_original')

    # Filter clusters for which more than a given proportion of cells are
    # assigned to a different cluster than that which they are most
    # correlated to
    cluster_annotation = cluster_annotation.query(
        'is_original_cluster_mean > {}'.format(is_original_cluster_mean_threshold))

    # Filter clusters smaller than a given size
    cluster_annotation = cluster_annotation.merge(
        cluster_annotation.groupby('cluster_id').size().rename('cluster_size').reset_index())
    cluster_annotation = cluster_annotation.query(
        'cluster_size >= {}'.format(cluster_size_threshold))

    # Assign clusters to s phase cells
    if metrics_data['is_s_phase'].any():
        # Assign s phase cells to the cluster they are most correlated with
        cell_filtered_clone_distances = cell_clone_distances.merge(
            cluster_annotation[['cluster_id']].drop_duplicates())
        s_phase_cluster = cell_filtered_clone_distances.merge(
            metrics_data.query('is_s_phase == True')[['cell_id']])
        s_phase_cluster = (
            s_phase_cluster
            .set_index(['cell_id', 'cluster_id'])['pearsonr']
            .unstack(level=['cluster_id']).idxmin(axis=1).rename('cluster_id').reset_index())

        # Filter s phase cells
        s_phase_filter = (filter_metrics
            .query('is_s_phase')
            .query('filter_reads')
            .query('filter_copy_state_diff')
            .query('filter_prop_hom_del'))[['cell_id']]
        s_phase_cluster = s_phase_cluster.merge(s_phase_filter)

        # Create a merged set of cluster calls
        final_clusters = pd.concat([
            cluster_annotation[['cell_id', 'cluster_id']],
            s_phase_cluster[['cell_id', 'cluster_id']],
        ], ignore_index=True)

    else:
        # No s-phase cells
        final_clusters = cluster_annotation[['cell_id', 'cluster_id']].copy()

    assert not final_clusters['cell_id'].duplicated().any()

    # Plotting
    #
    
    # Plot final clusters heatmap
    logging.info('plotting clusters to {}*'.format(results_prefix + 'final_'))
    plot_cn_data = cn_data.merge(
        final_clusters[['cell_id', 'cluster_id']])
    plot_clones(plot_cn_data, 'cluster_id', results_prefix + 'final_')

    # Plot s phase proportions
    metrics_data['is_s_phase'] = metrics_data['is_s_phase'].astype(bool)
    s_plot_data = (
        metrics_data
        .merge(final_clusters[['cell_id', 'cluster_id']].drop_duplicates())
        .groupby('cluster_id').agg({'is_s_phase': (np.sum, len, np.mean)}).reset_index())
    s_plot_data.columns = ['clone', 'sum', 'len', 'proportion']

    fig = plt.figure(figsize=(6, 4))
    ax = fig.add_subplot(211)
    seaborn.barplot(ax=ax, x='clone', y='proportion', data=s_plot_data, color='0.5')
    seaborn.despine()
    ax = fig.add_subplot(212)
    seaborn.barplot(ax=ax, x='clone', y='len', data=s_plot_data, color='0.5')
    seaborn.despine()
    plt.tight_layout()
    fig.savefig(results_prefix + 'clone_s_phase.pdf', bbox_inches='tight')

    return final_clusters


def recalculate_distances(distance_metric, distance_method, clone_cn_matrix, cell_cn_matrix):
    """ Recalculate distances to closest cluster using some metric.
    """

    logging.info('Calculating clone cell {} distance'.format(distance_metric))
    cell_clone_corr = {}
    for cluster_id in clone_cn_matrix.columns:
        logging.info('Calculating distance for clone {}'.format(cluster_id))
        cell_clone_corr[cluster_id] = cell_cn_matrix.corrwith(
            clone_cn_matrix[cluster_id], method=distance_method)

    distances = pd.DataFrame(cell_clone_corr)
    distances.columns.name = 'cluster_id'
    distances = distances.stack().rename(distance_metric).reset_index()

    return distances


def calculate_cell_clone_distances(cn_data, clusters, results_prefix):
    """ Calculate the distance to the closest clone for multiple metrics.
    """

    logging.info('Create clone copy number table')
    clone_cn_data = (
        cn_data
            .merge(clusters[['cell_id', 'cluster_id']].drop_duplicates())
            .groupby(['chr', 'start', 'end', 'cluster_id'], observed=True)
            .agg({'copy': np.mean, 'state': np.median})
            .reset_index()
    )
    clone_cn_data['state'] = clone_cn_data['state'].round().astype(int)

    logging.info('Create matrix of cn data for all cells')
    cell_cn_matrix = (
        cn_data
            .set_index(['chr', 'start', 'end', 'cell_id'])['copy']
            .unstack(level=['cell_id']).fillna(0)
    )

    logging.info('Create a matrix of cn data for filtered clones')
    clone_cn_matrix = (
        clone_cn_data
            .set_index(['chr', 'start', 'end', 'cluster_id'])['copy']
            .unstack(level=['cluster_id']).fillna(0)
    )

    pearsonr_distances = recalculate_distances(
        'pearsonr',
        lambda u, v: 1. - scipy.stats.pearsonr(u, v)[0],
        clone_cn_matrix,
        cell_cn_matrix,
    )

    spearmanr_distances = recalculate_distances(
        'spearmanr',
        lambda u, v: 1. - scipy.stats.spearmanr(u, v)[0],
        clone_cn_matrix,
        cell_cn_matrix,
    )

    cityblock_distances = recalculate_distances(
        'cityblock',
        scipy.spatial.distance.cityblock,
        clone_cn_matrix,
        cell_cn_matrix,
    )

    clone_cell_distances = pd.concat([
        pearsonr_distances.set_index(['cell_id', 'cluster_id']),
        spearmanr_distances.set_index(['cell_id', 'cluster_id']),
        cityblock_distances.set_index(['cell_id', 'cluster_id']),
    ], axis=1).reset_index()

    return clone_cell_distances


def plot_clones(cn_data, cluster_col, plots_prefix=None):
    plot_data = cn_data.copy()
    bin_filter = (plot_data['gc'] <= 0) | (plot_data['copy'].isnull())
    plot_data.loc[bin_filter, 'state'] = 0
    plot_data.loc[plot_data['copy'] > 5, 'copy'] = 5.
    plot_data.loc[plot_data['copy'] < 0, 'copy'] = 0.

    logging.info('Plotting cluster cn matrix')

    num_clusters = len(cn_data[cluster_col].unique())
    fig = plt.figure(figsize=(15, num_clusters/8))
    scgenome.cnplot.plot_cluster_cn_matrix(
        fig, plot_data, 'state', cluster_field_name=cluster_col)
    if plots_prefix is not None:
        fig.savefig(plots_prefix + 'clone_cn.pdf', bbox_inches='tight')

    logging.info('Plotting cell cn matrix')

    fig = plt.figure(figsize=(20, 30))
    matrix_data = scgenome.cnplot.plot_clustered_cell_cn_matrix_figure(
        fig, plot_data, 'copy', cluster_field_name=cluster_col, raw=True)
    if plots_prefix is not None:
        fig.savefig(plots_prefix + 'raw_cn.pdf', bbox_inches='tight')

    logging.info('Plotting cell cn matrix figure')

    fig = plt.figure(figsize=(20, 30))
    matrix_data = scgenome.cnplot.plot_clustered_cell_cn_matrix_figure(
        fig, plot_data, 'state', cluster_field_name=cluster_col)
    if plots_prefix is not None:
        fig.savefig(plots_prefix + 'cn_state.pdf', bbox_inches='tight')

    logging.info('Generating clone data')

    clone_cn_data = (
        cn_data
            .groupby(['chr', 'start', 'end', 'cluster_id'], observed=True)
            .agg({'copy': np.mean, 'state': np.median})
            .reset_index()
    )
    clone_cn_data['state'] = clone_cn_data['state'].astype(float).round().astype(int)

    logging.info('Plotting clone data')

    num_clusters = len(clone_cn_data['cluster_id'].unique())
    fig = plt.figure(figsize=(20, 4 * num_clusters))
    idx = 1
    for cluster_id, plot_data in clone_cn_data.groupby('cluster_id'):
        ax = fig.add_subplot(num_clusters, 1, idx)
        scgenome.cnplot.plot_cell_cn_profile(
            ax, plot_data, 'copy', 'state')
        ax.set_ylabel(f'Clone {cluster_id} Total CN')
        idx += 1
    if plots_prefix is not None:
        fig.savefig(plots_prefix + 'total_cn_profiles.pdf', bbox_inches='tight')


def calculate_mitotic_errors(
        cn_data,
        clusters,
        results_prefix,
        chromosome_state_diff_threshold=0.75,
        include_y_chrom=False,
    ):

    # Calculate raw and integer copy state per cluster
    clone_cn_data = (
    cn_data
        .merge(clusters)
        .groupby(['chr', 'start', 'end', 'cluster_id'], observed=True)
        .agg({'copy': np.mean, 'state': np.median})
        .reset_index()
    )
    clone_cn_data['state'] = clone_cn_data['state'].round().astype(int)

    # Create clone / cell cn table 
    clone_cell_cn = cn_data.merge(clusters[['cell_id', 'cluster_id']])
    clone_cell_cn = clone_cell_cn.merge(
        clone_cn_data[['chr', 'start', 'end', 'cluster_id', 'state']].rename(columns={'state': 'clone_cn'}),
        on=['chr', 'start', 'end', 'cluster_id'])
    clone_cell_cn['state_diff'] = clone_cell_cn['state'] - clone_cell_cn['clone_cn']
    clone_cell_cn['bin_size'] = clone_cell_cn['end'] - clone_cell_cn['start'] + 1

    # Calculate proportion of each chromosome for each cell that has
    # a different state from the clone (mean_diff)
    size_state_diff = clone_cell_cn.set_index('state_diff', append=True).groupby(['chr', 'cell_id', 'state_diff'], observed=True)['bin_size'].sum().rename('state_size').reset_index()
    size_total = clone_cell_cn.groupby(['chr', 'cell_id'])['bin_size'].sum().astype(float).rename('total_size').reset_index()
    mean_state_diff = size_state_diff.merge(size_total)
    mean_state_diff['mean_diff'] = mean_state_diff['state_size'] / mean_state_diff['total_size']

    # Filtering and threshold at 0.75 of a chromosome as a mis-segregation
    if not include_y_chrom:
        mean_state_diff = mean_state_diff[mean_state_diff['chr'] != 'Y']
    mean_state_diff = mean_state_diff[mean_state_diff['state_diff'] != 0]
    mean_state_diff = mean_state_diff[mean_state_diff['mean_diff'] > chromosome_state_diff_threshold]

    # Count per state diff
    fig = plt.figure(figsize=(3, 2))
    plot_data = mean_state_diff.groupby('state_diff').size().rename('count').reset_index()
    seaborn.barplot(x='state_diff', y='count', data=plot_data)
    fig.savefig(results_prefix + 'misseg_state_diff_counts.pdf', bbox_inches='tight')

    # Count per state diff
    fig = plt.figure(figsize=(4, 2))
    ax = fig.add_subplot(111)
    plot_data = mean_state_diff.groupby('cell_id').size().rename('chr_count').reset_index()
    plot_data = plot_data.groupby('chr_count').size().rename('cell_count').reset_index()
    plot_data['proportion'] = plot_data['cell_count'] / len(clone_cell_cn['cell_id'].unique())
    plot_data = plot_data.query('chr_count > 0')
    chr_counts = range(1, plot_data['chr_count'].max() + 1)
    seaborn.barplot(ax=ax, x='chr_count', y='proportion', data=plot_data, order=chr_counts, color='0.75')
    ax.set_xlabel('Num. chromosomes')
    ax.set_ylabel('Prop. cells')
    seaborn.despine(trim=True)
    fig.savefig(results_prefix + 'misseg_state_diff_proportions.pdf', bbox_inches='tight')

    # Count per chromosome
    chromosomes = [str(a) for a in range(1, 23)] + ['X']
    if include_y_chrom:
        chromosomes.append('Y')
    fig = plt.figure(figsize=(7, 2.5))
    ax = fig.add_subplot(111)
    plot_data = mean_state_diff.query('state_diff > 0').groupby('chr').size().rename('count').reset_index()
    seaborn.barplot(x='chr', y='count', data=plot_data, order=chromosomes, color=seaborn.desaturate('red', 0.75))
    plot_data = mean_state_diff.query('state_diff < 0').groupby('chr').size().rename('count').reset_index()
    plot_data['count'] = -plot_data['count']
    seaborn.barplot(x='chr', y='count', data=plot_data, order=chromosomes, color=seaborn.desaturate('blue', 0.75))
    ax.set_xlabel('Chromosome')
    ax.set_ylabel('Count')
    seaborn.despine(trim=True)
    ax.set_yticklabels([int(abs(a)) for a in ax.get_yticks()])
    fig.savefig(results_prefix + 'misseg_chr_counts.pdf', bbox_inches='tight')

    return mean_state_diff




import numpy as np
import scipy.cluster.hierarchy as sch
import scipy.spatial.distance as dst
import sklearn.preprocessing
import pandas as pd

from anndata import AnnData
from collections.abc import Iterable
from typing import Union

import scgenome.preprocessing.transform


def sort_cells(
        adata: AnnData,
        layer_name: Union[None, str, Iterable[Union[None,str]]]='copy',
        cell_ids: Iterable[str]=None,
        bin_ids: Iterable[str]=None,
        standarize: bool=False,
    ) -> AnnData:
    """ Sort cells by hierarchical clustering on copy number values.

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data to use for sorting, None for X, by default 'copy'
    cell_ids : str, optional
        subset of cells to cluster, by default None
    bin_ids : str, optional
        subset of bins to cluster, by default None
    standarize : bool
        standardize the data prior to sorting, by default False

    Returns
    -------
    AnnData
        copy number data with cell_order column added to obs
    """
    if cell_ids is None:
        cell_ids = adata.obs.index

    if bin_ids is None:
        bin_ids = adata.var.index

    def __get_layer(layer_name):
        if layer_name is not None:
            return np.array(adata[cell_ids, bin_ids].layers[layer_name])
        else:
            return np.array(adata[cell_ids, bin_ids].X)

    if isinstance(layer_name, (str, type(None))):
        X = __get_layer(layer_name)
    elif isinstance(layer_name, Iterable):
        X = np.concatenate([__get_layer(l) for l in layer_name], axis=1)
    else:
        raise ValueError(f'layer_name was {layer_name}')

    X = scgenome.preprocessing.transform.fill_missing(X)

    if standarize:
        X = sklearn.preprocessing.StandardScaler().fit_transform(X)

    D = dst.squareform(dst.pdist(X, 'cityblock'))
    Y = sch.linkage(D, method='complete')
    Z = sch.dendrogram(Y, color_threshold=-1, no_plot=True)
    idx = np.array(Z['leaves'])

    ordering = np.zeros(idx.shape[0], dtype=int)
    ordering[idx] = np.arange(idx.shape[0])

    adata.obs['cell_order'] = '-1'
    adata.obs.loc[cell_ids, 'cell_order'] = pd.Series(ordering, index=adata.obs.loc[cell_ids].index)

    return adata




import collections
import os
import packaging.version
import yaml

import scgenome.utils


def find_filenames(filenames, suffix):
    return [f for f in filenames if f.endswith(suffix)]


def _find_manifest_filenames(results_dir):
    for dirpath, dirnames, filenames in os.walk(results_dir):
        for filename in filenames:
            if not filename == 'metadata.yaml':
                continue

            manifest_filename = os.path.join(dirpath, filename)

            yield manifest_filename


def find_results_directories(results_dir):
    results_directories = collections.defaultdict(list)

    for manifest_filename in _find_manifest_filenames(results_dir):
        manifest = yaml.safe_load(open(manifest_filename))

        # Kludge: mondrian v1 has name instead of type
        if 'type' not in manifest['meta']:
            manifest['meta']['type'] = manifest['meta']['name']

        results_type = manifest['meta']['type']

        # KLUDGE: 0.3.1 -> v0.3.1
        if not manifest['meta']['version'].startswith('v'):
            manifest['meta']['version'] = 'v' + manifest['meta']['version']

        results_directories[results_type].append(os.path.dirname(manifest_filename))

    return results_directories


def find_results_directory(results_dir, analysis_type):
    analysis_results_dirs = find_results_directories(results_dir)

    if analysis_type not in analysis_results_dirs:
        raise Exception(f'no analysis type {analysis_type} found at {results_dir}')

    if len(analysis_results_dirs[analysis_type]) > 1:
        raise Exception(f'found {len(analysis_results_dirs[analysis_type])} analyses of type {analysis_type} found at {results_dir}')

    return analysis_results_dirs[analysis_type][0]


def get_version(results_dir):
    """ Get version for a given results.
    
    Args:
        results_dir (str): pseudobulk results directory
    
    Returns:
        str: results version
    """
    manifest_filename = os.path.join(results_dir, 'metadata.yaml')
    manifest = yaml.safe_load(open(manifest_filename))
    return manifest['meta']['version']


def find_results_filepath(results_dir, filename_suffix, result_type=None, analysis_type=None):
    """ Get filepaths for libraries and samples by suffix
    
    Args:
        results_dir (str): pseudobulk results directory
        filename_suffix (str): suffix of requested files
        result_type (str): requested result type

    KwArgs:
        analysis_type (str): check analysis type

    Returns:
        str: filepath
    """

    if analysis_type is not None:
        results_dir = find_results_directory(results_dir, analysis_type)

    manifest_filename = os.path.join(results_dir, 'metadata.yaml')
    manifest = yaml.safe_load(open(manifest_filename))

    # Kludge: mondrian v1 has name instead of type
    if 'type' not in manifest['meta']:
        manifest['meta']['type'] = manifest['meta']['name']

    if analysis_type is not None and manifest['meta']['type'] != analysis_type:
        raise Exception(f"expected analysis {analysis_type} and found {manifest['meta']['type']}")

    # Two formats:
    #  - filenames for just the names of the files,
    #  - files for file information keyed by filename.
    if 'filenames' in manifest:
        filenames = list(filter(lambda a: a.endswith(filename_suffix), manifest['filenames']))

        if len(filenames) != 1:
            raise Exception(f'found {len(filenames)} {filename_suffix} files for {results_dir}: {filenames}')

        filename = filenames[0]
        filepath = os.path.join(results_dir, filename)

    elif 'files' in manifest:
        assert result_type is not None

        files = list(filter(lambda a: a[1]['result_type'] == result_type and not a[0].endswith('.yaml'), manifest['files'].items()))

        if len(files) != 1:
            raise Exception(f'found {len(files)} {filename_suffix} files for {results_dir}: {files}')

        filename = files[0][0]
        filepath = os.path.join(results_dir, filename)

    else:
        raise ValueError()

    return filepath


def get_pseudobulk_files(results_dir, suffix):
    """ Get files for libraries and samples by suffix
    
    Args:
        results_dir (str): pseudobulk results directory
        suffix (str): suffix of requested files
    
    Yields:
        (str, str, str): sample id, library id, filename
    """

    manifest_filename = os.path.join(results_dir, 'metadata.yaml')
    manifest = yaml.safe_load(open(manifest_filename))

    if packaging.version.parse(manifest['meta']['version']) < packaging.version.parse('v0.5.0'):
        for a in _get_pseudobulk_files_v_lt_050(results_dir, suffix):
            yield a
        return

    filenames = list(filter(lambda a: a.endswith(suffix), manifest['filenames']))

    if len(filenames) != 1:
        raise ValueError(f'found {len(filenames)} {suffix} files for {results_dir}: {filenames}')

    filename = filenames[0]
    filepath = os.path.join(results_dir, filename)

    yield None, None, filepath


def _get_pseudobulk_files_v_lt_050(results_dir, suffix):
    manifest_filename = os.path.join(results_dir, 'metadata.yaml')
    manifest = yaml.safe_load(open(manifest_filename))

    tumour_samples = manifest['meta']['tumour_samples']
    filenames = manifest['filenames']

    for sample_info in tumour_samples:
        sample_id = sample_info['sample_id']
        library_id = sample_info['library_id']

        sample_lib_suffix = f'{sample_id}_{library_id}_{suffix}'
        sample_lib_filenames = list(filter(lambda a: a.endswith(sample_lib_suffix), filenames))

        if len(sample_lib_filenames) != 1:
            raise ValueError(
                f'found {len(sample_lib_filenames)} {suffix} files for {sample_id}, {library_id}, {results_dir}')

        sample_lib_filename = sample_lib_filenames[0]
        sample_lib_filepath = os.path.join(results_dir, sample_lib_filename)

        yield sample_id, library_id, sample_lib_filepath


def concat_results(results):
    """ Merge a set of results tables

    Args:
        tables (list of dict of DataFrame): list of named results tables

    Returns:
        dict of DataFrame: named results tables with unified categories
    """

    names = set()
    for tables in results:
        for name in tables.keys():
            names.add(name)

    concatenated = {}
    for name in names:
        concatenated[name] = []
        for tables in results:
            if name in tables:
                concatenated[name].append(tables[name])

    for name in names:
        concatenated[name] = scgenome.utils.concat_with_categories(concatenated[name], ignore_index=True)

    return concatenated



import logging
import pandas as pd
import numpy as np
import anndata as ad
import pyranges as pr

import scgenome.loaders.qc
import csverve

from anndata import AnnData
from pyranges import PyRanges
from typing import Dict, Sequence
from pandas import DataFrame


def read_dlp_hmmcopy(alignment_results_dir, hmmcopy_results_dir, annotation_results_dir, sample_ids=None, additional_hmmcopy_reads_cols=None) -> AnnData:
    """ Read hmmcopy results from the DLP pipeline.

    Parameters
    ----------
    alignment_results_dir : str
        dlp pipeline alignment results directory
    hmmcopy_results_dir : str
        dlp pipeline hmmcopy results directory
    annotation_results_dir : str
        dlp pipeline annotation results directory
    sample_ids : list
        sample ids to load
    additional_hmmcopy_reads_cols : list
        per bin metrics to load

    Returns
    -------
    AnnData
        An instantiated AnnData Object.
    """


    results = scgenome.loaders.qc.load_qc_results(
        alignment_results_dir,
        hmmcopy_results_dir,
        annotation_results_dir,
        sample_ids=sample_ids,
        additional_hmmcopy_reads_cols=additional_hmmcopy_reads_cols,
    )

    metrics_data = results['annotation_metrics']
    cn_data = results['hmmcopy_reads']

    return convert_dlp_hmmcopy(metrics_data, cn_data)


def read_dlp_hmmcopy2(reads_filename, metrics_filename, sample_ids=None) -> AnnData:
    """ Read hmmcopy results from the DLP pipeline.

    Parameters
    ----------
    reads_filename (str):
        dlp pipeline reads filename
    metrics_filename (str):
        dlp pipeline metrics filename
    sample_ids (list):
        sample ids to load

    Returns
    -------
    AnnData
        An instantiated AnnData Object.
    """

    cn_data = csverve.read_csv(
        reads_filename,
        dtype={
            'cell_id': 'category',
            'sample_id': 'category',
            'library_id': 'category',
            'chr': 'category',
        })

    metrics_data = csverve.read_csv(
        metrics_filename,
        dtype={
            'cell_id': 'category',
            'sample_id': 'category',
            'library_id': 'category',
        })

    scgenome.utils.union_categories([cn_data, metrics_data])

    return convert_dlp_hmmcopy(metrics_data, cn_data)


def create_cn_anndata(
        cn_data: DataFrame,
        X_column: str,
        layers_columns: Sequence[str],
        cell_metrics_data: DataFrame=None,
        bin_metrics_data: DataFrame=None,
    ) -> AnnData:
    """ Convert hmmcopy pandas dataframes to anndata

    Parameters
    ----------
    cn_data : DataFrame
        copy number data per cell in long format
    X_column : str
        column of cn_data to use for X
    layers_columns : Sequence[str]
        columns of cn_data to use for lauers
    cell_metrics_data : DataFrame, optional
        per cell metrics data, by default None
    bin_metrics_data : DataFrame, optional
        per bin metrics data, by default None

    Returns
    -------
    AnnData
        An instantiated AnnData Object.

    Raises
    ------
    ValueError
        duplicate data or otherwise incompatible inputs
    """
    if cell_metrics_data is None:
        cell_metrics_data = cn_data[['cell_id']].drop_duplicates()

    if bin_metrics_data is None:
        bin_metrics_data = cn_data[['chr', 'start', 'end']].drop_duplicates()

    duplicate_cell_ids = cn_data.loc[cn_data[['chr', 'start', 'end', 'cell_id']].duplicated(keep=False), 'cell_id'].unique()
    if len(duplicate_cell_ids) > 0:
        raise ValueError(f'cell {duplicate_cell_ids[0]} is duplicated, and {len(duplicate_cell_ids)} others')

    assert not cell_metrics_data.duplicated(subset=['cell_id']).any()
    assert not bin_metrics_data.duplicated(subset=['chr', 'start', 'end']).any()

    X = (
        cn_data
            .set_index(['chr', 'start', 'end', 'cell_id'])[X_column]
            .unstack(level='cell_id')
            .transpose())
    X.index = X.index.astype(str)

    chr_start_end_index = X.columns
    bin_index = pd.Series(
        X.columns.get_level_values('chr').astype(str) + ':' +
        X.columns.get_level_values('start').astype(str) + '-' +
        X.columns.get_level_values('end').astype(str),
        name='bin')

    X = X.set_axis(bin_index, axis=1, copy=False)

    layers = {}
    for layer_name in layers_columns:
        layers[layer_name] = (
            cn_data
                .set_index(['chr', 'start', 'end', 'cell_id'])[layer_name]
                .unstack(level='cell_id')
                .transpose()
                .reindex(index=X.index, columns=chr_start_end_index)
                .set_axis(bin_index, axis=1, copy=False))
        layers[layer_name].index = layers[layer_name].index.astype(str)

    bin_data = (
        bin_metrics_data
            .set_index(['chr', 'start', 'end'], drop=False)
            .reindex(index=chr_start_end_index)
            .set_axis(bin_index, axis=0))

    cell_data = (
        cell_metrics_data
            .set_index(['cell_id'])
            .reindex(X.index))
    cell_data.index = cell_data.index.astype(str)

    adata = ad.AnnData(
        X,
        dtype=X.dtypes[0],
        obs=cell_data,
        var=bin_data,
        layers=layers,
    )

    return adata


def convert_dlp_hmmcopy(metrics_data: DataFrame, cn_data: DataFrame) -> AnnData:
    """ Convert hmmcopy pandas dataframes to anndata

    Parameters
    ----------
    metrics_data : DataFrame
        hmmcopy metrics
    cn_data : DataFrame
        hmmcopy reads data

    Returns
    -------
    AnnData
        An instantiated AnnData Object.
    """

    return create_cn_anndata(
        cn_data[['cell_id', 'chr', 'start', 'end', 'reads', 'copy', 'state']],
        layers_columns = ['copy', 'state'],
        X_column = 'reads',
        cell_metrics_data=metrics_data,
        bin_metrics_data=cn_data[['chr', 'start', 'end', 'gc']].drop_duplicates(subset=['chr', 'start', 'end']))


def convert_dlp_signals(hscn: DataFrame, metrics_data: DataFrame) -> AnnData:
    """ Convert signals pandas dataframes to anndata

    Parameters
    ----------
    hscn : DataFrame
        signals reads data

    Returns
    -------
    AnnData
        An instantiated AnnData Object.
    """
    hscn['state_a'] = hscn['state_AS_phased'].str.split('|', expand=True)[0].astype(float)
    hscn['state_b'] = hscn['state_AS_phased'].str.split('|', expand=True)[1].astype(float)

    layers_columns = [
        'copy', 'state',
        'alleleA', 'alleleB', 'BAF',
        'Maj', 'Min',
        'state_a', 'state_b',
    ]

    return create_cn_anndata(
        hscn,
        layers_columns=layers_columns,
        X_column='totalcounts',
        cell_metrics_data=metrics_data,
    )


def _convert_pyranges(r):
    return r.df.rename(columns={
        'Chromosome': 'chr',
        'Start': 'start',
        'End': 'end',
    })


def _add_bin_index(df):
    df['bin'] = (
        df['chr'].astype(str) + ':' +
        df['start'].astype(str) + '-' +
        df['end'].astype(str))
    return df.set_index('bin')
    

def read_bam_bin_counts(bins: PyRanges, bams: Dict[str, str], excluded: PyRanges = None, **kwargs) -> AnnData:
    """ Count reads in bins from bams

    Parameters
    ----------
    bins : pyranges.PyRanges
        bins in which to count reads
    bams : Dict[str]
        bam filenames with cell ids as keys
    excluded: PyRanges
        excluded genomic regions to filter reads

    Returns
    -------
    ad.AnnData
        binned read counts
    """

    bin_data = _convert_pyranges(bins)
    bin_data = _add_bin_index(bin_data)

    cn_matrix = {}

    for cell_id, cell_bam in bams.items():
        logging.info(f"reading {cell_bam}")
        bam_data = pr.read_bam(cell_bam, **kwargs)

        if excluded is not None:
            logging.info("excluding reads")
            bam_data = bam_data.intersect(excluded, invert=True)

        logging.info(f"count overlaps")
        bam_data = bam_data.intersect(bins, how='containment')
        read_counts = bins.count_overlaps(bam_data, overlap_col='reads')

        read_counts = _convert_pyranges(read_counts)
        read_counts = _add_bin_index(read_counts)

        assert (read_counts.index == bin_data.index).all()

        cn_matrix[cell_id] = read_counts['reads'].values.copy()

    cn_matrix = pd.DataFrame(cn_matrix, index=bin_data.index)

    cell_data = pd.DataFrame({'cell_id': cn_matrix.columns.values}).set_index('cell_id')

    adata = ad.AnnData(
        cn_matrix.T,
        obs=cell_data,
        var=bin_data,
    )

    return adata


def read_medicc2_cn(cn_profiles_filename, allele_specific: bool = False) -> AnnData:
    """ Read medicc2 results

    Parameters
    ----------
    cn_profiles_filename : str
        Copy number profiles filename
    allele_specific : bool, optional
        _description_, by default False

    Returns
    -------
    AnnData
        Medicc CN results.
    """

    cn_data = pd.read_csv(
        cn_profiles_filename,
        sep='\t',
        dtype={
            'chrom': 'category',
            'sample_id': 'category'
    })

    cn_data = cn_data.rename(columns={
        'sample_id': 'cell_id',
        'chrom': 'chr',
    })

    if allele_specific:
        cn_data['state'] = cn_data['cn_a'] + cn_data['cn_b']
        cn_fields = ['cn_a', 'cn_b']

    else:
        cn_data = cn_data.rename(columns={'cn': 'state'})
        cn_fields = []

    cn_data['chr'] = cn_data['chr'].str.replace('chr', '')
    cn_data.loc[cn_data['chr'] == '23', 'chr'] = 'X'
    cn_data.loc[cn_data['chr'] == '24', 'chr'] = 'Y'
    cn_data['chr'] = cn_data['chr'].astype('category')

    cn_data['bin'] = cn_data['chr'].astype(str) + ':' + cn_data['start'].astype(str) + '-' + cn_data['end'].astype(str)

    cn_matrix = (
        cn_data
            .set_index(['bin', 'cell_id'])[cn_fields + ['state', 'is_gain', 'is_loss']]
            .unstack(level='cell_id')
            .transpose())

    bin_data = (
        cn_data[['bin', 'chr', 'start', 'end', 'is_normal', 'is_clonal']]
            .drop_duplicates(subset=['bin'])
            .set_index(['bin'])
            .reindex(cn_matrix.loc['state'].columns))

    cell_data = (
        cn_data[['cell_id', 'is_wgd']]
            .drop_duplicates()
            .set_index(['cell_id'])
            .reindex(cn_matrix.loc['state'].index))

    cell_data['is_root'] = cell_data.index == 'diploid'
    cell_data['is_internal'] = cell_data.index.to_series().str.startswith('internal_')
    cell_data['is_cell'] = (~cell_data['is_root']) & (~cell_data['is_internal'])

    X = cn_matrix.loc['state'].astype(np.float32)

    layers = {
        'is_gain': cn_matrix.loc['is_gain'],
        'is_loss': cn_matrix.loc['is_loss'],
    }

    for field in cn_fields:
        layers[field] = cn_matrix.loc[field]

    X.index = X.index.astype(str)
    cell_data.index = cell_data.index.astype(str)
    for field in layers:
        layers[field].index = layers[field].index.astype(str)

    adata = ad.AnnData(
        X,
        obs=cell_data,
        var=bin_data,
        layers=layers,
    )

    return adata





def plot_gc_reads(adata, obs_id, *args, **kwargs):
    """ Plot scatter points of gc by read count.

    Args:
        adata (anndata.AnnData): copy number
        obs_id (str): cell or clone to plot
    """

    raise NotImplementedError



import matplotlib
import seaborn
import logging
import scipy.special
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from hmmlearn._hmmc import _viterbi
from scipy.stats import binom
from matplotlib import collections  as mc

import scgenome.refgenome
import scgenome.utils
import scgenome.cnplot


# TODO: possibly deprecated
def get_clone_snp_count(het_counts, cluster_df):
    cell_ids = het_counts['cell_ids']
    cell_ids.set_index('cell_id', inplace=True)

    snp_data = None

    cluster_df = cluster_df.merge(cell_ids.reset_index()[['cell_id']].drop_duplicates())

    print(cluster_df.shape[0])
    for cell_id, cluster_id in cluster_df[['cell_id', 'cluster_id']].values:
        print('.', end=' ')
        cell_index = cell_ids.loc[cell_id, 'cell_index']
        df = het_counts.select('/allele_counts', where='cell_index=={}'.format(cell_index))
        df['cluster_id'] = cluster_id
        df['chromosome'] = df['chromosome'].astype(str)
        if snp_data is None:
            snp_data = df
        else:
            snp_data = pd.concat([snp_data, df], ignore_index=True)
            snp_data = snp_data.groupby(['chromosome', 'coord', 'cluster_id'], observed=True)[['ref_counts', 'alt_counts']].sum().reset_index()

    snp_data['total_counts'] = snp_data['ref_counts'] + snp_data['alt_counts']
    snp_data['vaf'] = snp_data['alt_counts'] / snp_data['total_counts'].astype(float)

    return snp_data


# TODO: possibly deprecated
def calculate_haplotype_allele_counts(snp_data, haps, bin_size):
    haps['swap'] = (haps['allele'] == haps['allele_id'])
    haps = haps[['chromosome', 'position', 'hap_label', 'swap']].drop_duplicates().rename(columns={'position': 'coord'})

    data = snp_data.merge(haps)
    data['allele_1'] = np.where(data['swap'], data['ref_counts'], data['alt_counts'])
    data['allele_2'] = np.where(~data['swap'], data['ref_counts'], data['alt_counts'])
    data['bin'] = (data['coord'] / bin_size).astype(int)

    # Aggregated by bin and haplotype block
    data2 = (
        data.groupby(['cluster_id', 'chromosome', 'hap_label', 'bin'], observed=True)[['allele_1', 'allele_2', 'total_counts', 'coord']]
        .aggregate({'allele_1': sum, 'allele_2': sum, 'total_counts': sum, 'coord': (np.min, np.max, len)}).reset_index())
    data2.columns = ['_'.join(col).strip().rstrip('_') for col in data2.columns.values]
    data2['maf'] = np.minimum(data2['allele_1_sum'], data2['allele_2_sum']) / data2['total_counts_sum'].astype(float)
    data2['start'] = data2['bin'] * bin_size + 1
    data2['end'] = (data2['bin'] + 1) * bin_size

    hap_data = data2[[
        'chromosome', 'start', 'end', 'hap_label',
        'cluster_id', 'allele_1_sum', 'allele_2_sum', 'total_counts_sum'
    ]].rename(columns={'chromosome': 'chr'})

    return hap_data


def plot_vaf_cn_profile(ax, hap_data, allele_cn):
    """ Plot genome wide VAF and predicted CN state for haplotype alleles.
    """
    allele_cn = allele_cn.query('total_cn != 0')

    chrom_info = plot_vaf_profile(
        ax, hap_data, 'maf',
        size_field_name='total_counts_sum',
        size_scale=100.,
    )

    allele_cn['cn_ratio'] = allele_cn['minor_cn'] / allele_cn['total_cn']

    allele_cn = allele_cn[allele_cn['chr'].isin(scgenome.refgenome.info.chromosomes)]

    allele_cn.set_index('chr', inplace=True)
    allele_cn['chromosome_start'] = scgenome.refgenome.info.chromosome_start
    allele_cn.reset_index(inplace=True)

    allele_cn['start'] = allele_cn['start'] + allele_cn['chromosome_start']
    allele_cn['end'] = allele_cn['end'] + allele_cn['chromosome_start']

    lines = np.zeros((allele_cn.shape[0], 2, 2))
    lines[:, 0, 0] = allele_cn['start'].values
    lines[:, 1, 0] = allele_cn['end'].values
    lines[:, 0, 1] = allele_cn['cn_ratio'].values
    lines[:, 1, 1] = allele_cn['cn_ratio'].values

    lc = mc.LineCollection(lines, colors='b', linewidths=1.)
    ax.add_collection(lc)

    return chrom_info


def plot_vaf_profile(ax, cn_data, value_field_name, cn_field_name=None, size_field_name=None, size_scale=1.):
    """ Plot genome wide VAF profile for haplotype alleles
    """
    plot_data = cn_data.copy()
    plot_data = plot_data[plot_data['chr'].isin(scgenome.refgenome.info.chromosomes)]

    plot_data.set_index('chr', inplace=True)
    plot_data['chromosome_start'] = scgenome.refgenome.info.chromosome_start
    plot_data.reset_index(inplace=True)

    plot_data['start'] = plot_data['start'] + plot_data['chromosome_start']
    plot_data['end'] = plot_data['end'] + plot_data['chromosome_start']

    c = '0.75'
    cmap = None
    if cn_field_name is not None:
        c = plot_data[cn_field_name]
        cmap = scgenome.cnplot.get_cn_cmap(plot_data[cn_field_name].values)

    s = 1
    if size_field_name is not None:
        s = plot_data[size_field_name] / size_scale

    ax.scatter(
        plot_data['start'], plot_data[value_field_name], c=c, s=s, alpha=0.1, cmap=cmap)

    ax.set_xlim((-0.5, scgenome.refgenome.info.chromosome_end.max()))
    ax.set_xlabel('chromosome')
    ax.set_xticks([0] + list(scgenome.refgenome.info.chromosome_end.values))
    ax.set_xticklabels([])
    ax.set_yticks([0., 0.1, 0.2, 0.3, 0.4, 0.5])
    ax.xaxis.tick_bottom()
    ax.yaxis.tick_left()
    ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(scgenome.refgenome.info.chromosome_mid))
    ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter(scgenome.refgenome.info.chromosomes))

    seaborn.despine(ax=ax, offset=10, trim=True)

    chrom_info = (
        plot_data[['chr', 'chromosome_start']]
        .drop_duplicates()
        .set_index('chr')['chromosome_start'])

    return chrom_info


def infer_allele_cn(clone_cn_data, hap_data, loh_error_rate=0.01):
    """ HMM inference of clone and allele specific copy number based on haplotype
    allele read counts.
    """
    cn = clone_cn_data.merge(
        hap_data,
        on=['chr', 'start', 'end', 'cluster_id'],
        how='left',
    ).fillna(0).rename(columns={'integer_copy_number': 'total_cn'})

    cn['total_cn'] = cn['total_cn'].astype(int).astype(float)

    total_cn = cn['total_cn'].values

    minor_cn = np.tile(np.arange(0, 10, 1), (cn.shape[0], 1))
    n = np.tile(cn['total_counts_sum'].astype(int), (minor_cn.shape[1], 1)).T
    x = np.tile(cn[['allele_1_sum', 'allele_2_sum']].min(axis=1).astype(int), (minor_cn.shape[1], 1)).T
    p = minor_cn / total_cn[:, np.newaxis]

    # Add an error term for the loh state
    # Rational: we can think of each states probability of a minor allele
    # read as being beta distributed around the expected calculated as
    # the ratio of minor to total copy number.  For most states the expectation
    # will be equal to the copy number ratio except for the loh state which will
    # have the configurable offset from 0 governed by the loh_error_rate term
    p[:, 0] = loh_error_rate

    l = binom.logpmf(x, n, p)
    l = np.logaddexp(l, np.log(0.01))

    # Set the likelihood to be very low for impossible states
    l[minor_cn > total_cn[:, np.newaxis]] = -1000.

    # Set the likelihood to a low value that is still greater than the
    # impossible state value if the likelihood is nan.  This will primarily
    # catch the case where minor copy number equals total copy number, which
    # is in general not a valid solution unless total copy number is 0
    l[np.isnan(l)] = -100.

    minor_cn_data = (pd.DataFrame(l, index=cn.set_index(['cluster_id', 'chr', 'start', 'end', 'hap_label']).index)
        .groupby(level=[0, 1, 2, 3]).sum())

    minor_cn = pd.DataFrame(index=minor_cn_data.index)
    minor_cn['minor_cn'] = None

    for clone_id in cn['cluster_id'].unique():
        framelogprob = minor_cn_data.loc[clone_id].values

        N_chain = framelogprob.shape[0]
        N_states = framelogprob.shape[1]

        seq, prob = _viterbi(
            N_chain,
            N_states,
            np.log(np.ones(N_states) / N_states),
            np.log(np.eye(N_states) * 1e4 + np.ones((N_states, N_states))),
            framelogprob)

        minor_cn.loc[clone_id, 'minor_cn'] = seq

    minor_cn = minor_cn.reset_index()
    minor_cn = minor_cn.merge(
        cn[['chr', 'start', 'end', 'cluster_id', 'total_cn']].drop_duplicates(),
        how='left')
    minor_cn['total_cn'] = minor_cn['total_cn'].astype(int)
    minor_cn['major_cn'] = minor_cn['total_cn'] - minor_cn['minor_cn']

    assert minor_cn['minor_cn'].notnull().any()
    assert minor_cn['major_cn'].notnull().any()
    assert minor_cn['total_cn'].notnull().any()

    return minor_cn


def calculate_cluster_allele_counts(allele_data, clusters, cn_bin_size):
    """ Calculate allele specific haplotype allele counts per cluster
    """
    index_cols = [
        'chromosome',
        'start',
        'end',
        'hap_label',
    ]

    # Create allele 1 and 2 counts matrix
    allele_data = allele_data.set_index(index_cols + ['cell_id', 'allele_id'])['readcount'].unstack(fill_value=0)
    assert '0' in allele_data.columns and '1' in allele_data.columns
    allele_data.rename(columns={'0': 'allele_1', '1': 'allele_2'}, inplace=True)
    allele_data.reset_index(inplace=True)

    # Merge clusters and redo categoricals
    scgenome.utils.union_categories(
        [allele_data, clusters],
        cat_cols=['cell_id'])
    allele_data = allele_data.merge(clusters[['cell_id', 'cluster_id']])
    allele_data = allele_data.groupby(
        index_cols + ['cluster_id'],
        as_index=True, observed=True)[['allele_1', 'allele_2']].sum().reset_index()

    allele_data['total'] = allele_data['allele_1'] + allele_data['allele_2']
    allele_data['start'] = (allele_data['start'] / cn_bin_size).astype(int) * cn_bin_size + 1
    allele_data['end'] = allele_data['start'] + cn_bin_size - 1

    return allele_data


def calculate_cluster_allele_cn(
        cn_data, allele_data, clusters,
        total_allele_counts_threshold=6,
        plots_prefix=None,
    ):
    """ Infer allele and cluster specific copy number from haplotype allele counts
    """
    clone_cn_state = (
        cn_data.merge(clusters[['cell_id', 'cluster_id']])
        .groupby(['chr', 'start', 'end', 'cluster_id'], observed=True)['state']
        .median().astype(int).reset_index())

    clone_cn_copy = (
        cn_data.merge(clusters[['cell_id', 'cluster_id']])
        .groupby(['chr', 'start', 'end', 'cluster_id'], observed=True)['copy']
        .mean().reset_index())

    clone_cn_data = clone_cn_state.merge(clone_cn_copy)

    clone_cn_data['total_cn'] = clone_cn_data['state']

    allele_data = allele_data.rename(columns={
        'chromosome': 'chr',
        'total': 'total_counts_sum',
        'allele_1': 'allele_1_sum',
        'allele_2': 'allele_2_sum',
    })

    allele_data = allele_data[allele_data['total_counts_sum'] >= total_allele_counts_threshold]

    allele_cn = scgenome.snpdata.infer_allele_cn(clone_cn_data, allele_data)

    allele_data['maf'] = (
        np.minimum(allele_data['allele_1_sum'], allele_data['allele_2_sum']) /
        allele_data['total_counts_sum'].astype(float))

    num_clusters = len(clusters['cluster_id'].unique())
    fig = plt.figure(figsize=(20, 4 * num_clusters))
    idx = 1
    for cluster_id in clusters['cluster_id'].unique():
        cluster_allele_data = allele_data.query('cluster_id == {}'.format(cluster_id))
        cluster_allele_cn = allele_cn.query('cluster_id == {}'.format(cluster_id))

        ax = fig.add_subplot(num_clusters, 1, idx)
        scgenome.snpdata.plot_vaf_cn_profile(
            ax, cluster_allele_data, cluster_allele_cn)
        ax.set_ylabel(f'Clone {cluster_id} BAF')

        idx += 1

    if plots_prefix is not None:
        fig.savefig(plots_prefix + 'allele_cn_profiles.png', bbox_inches='tight')

    return allele_cn




import logging
import os
import shutil
import gzip

import numpy as np
import pandas as pd
import yaml


class getFileHandle(object):
    def __init__(self, filename, mode='rt'):
        self.filename = filename
        self.mode = mode

    def __enter__(self):
        if self.get_file_format(self.filename) in ["csv", 'plain-text', "yaml"]:
            self.handle = open(self.filename, self.mode)
        elif self.get_file_format(self.filename) == "gzip":
            self.handle = gzip.open(self.filename, self.mode)
        elif self.get_file_format(self.filename) == "h5":
            self.handle = pd.HDFStore(self.filename, self.mode)
        return self.handle

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.handle.close()

    def get_file_format(self, filepath):
        if filepath.endswith('.tmp'):
            filepath = filepath[:-4]

        _, ext = os.path.splitext(filepath)

        if ext == ".csv":
            return "csv"
        elif ext == ".gz":
            return "gzip"
        elif ext == ".h5" or ext == ".hdf5":
            return "h5"
        elif ext == ".yaml":
            return "yaml"
        else:
            logging.getLogger("single_cell.helpers").warn(
                "Couldn't detect output format for {}. extension {}".format(filepath, ext)
            )
            return "plain-text"


class CsvParseError(Exception):
    pass

class CsvInputError(Exception):
    pass

class CsvWriterError(Exception):
    pass


def get_dtypes_from_df(df, na_rep='NA'):
    pandas_to_std_types = {
        "bool": "bool",
        "int64": "int",
        "float64": "float",
        "object": "str",
        "category": "str",
    }

    typeinfo = {}
    for column, dtype in df.dtypes.items():
        if column in ['chr', 'chrom', 'chromosome']:
            typeinfo[column] = 'str'
        else:
            if df.empty:
                typeinfo[column] = na_rep
            else:
                typeinfo[column] = pandas_to_std_types[str(dtype)]

    return typeinfo


class CsvInput(object):
    def __init__(self, filepath, na_rep='NA'):
        """
        csv file and all related metadata
        :param filepath: path to csv
        :type filepath: str
        :param na_rep: replace na with this
        :type na_rep: str
        """
        self.filepath = filepath
        self.compression = self.__get_compression_type_pandas()
        self.na_rep = na_rep

        if os.path.exists(self.yaml_file):
            metadata = self.__parse_metadata()
        else:
            metadata = self.generate_metadata()

        self.header, self.sep, self.dtypes, self.columns = metadata

    @property
    def yaml_file(self):
        return self.filepath + '.yaml'

    def __detect_sep_from_header(self, header):
        """
        detect whether file is tab or comma separated from header
        :param header: header line
        :type header: str
        :return: separator
        :rtype: str
        """
        if '\t' in header and ',' in header:
            raise CsvParseError("Unable to detect separator from {}".format(header))

        if '\t' not in header and ',' not in header:
            raise CsvParseError("Unable to detect separator from {}".format(header))

        if '\t' in header:
            return '\t'
        elif ',' in header:
            return ','

    def __detect_sep_from_file(self):
        with getFileHandle(self.filepath) as reader:
            header = reader.readline().strip()
            return self.__detect_sep_from_header(header)

    def __get_compression_type_pandas(self):
        filepath = self.filepath
        if filepath.endswith('.tmp'):
            filepath = filepath[:-4]

        _, ext = os.path.splitext(filepath)

        if ext == ".csv":
            return None
        elif ext == ".gz":
            return "gzip"
        elif ext == ".h5" or ext == ".hdf5":
            raise CsvInputError("HDF is not supported")
        else:
            logging.getLogger("single_cell.utils.csv").warn(
                "Couldn't detect output format for {}. extension {}".format(filepath, ext)
            )
            return None

    def __is_empty(self):
        with open(self.filepath) as f:
            first_line = f.readline()
            if len(first_line) == 0:
                logging.getLogger("single_cell.utils.csv").warn("empty csv file")
                return True

    def __generate_dtypes(self, columns=None, sep=','):
        data = pd.read_csv(
            self.filepath, compression=self.compression, chunksize=10 ** 6,
            sep=sep
        )
        data = next(data)

        if columns:
            data.columns = columns

        typeinfo = get_dtypes_from_df(data)
        return typeinfo

    def __parse_metadata(self):
        with getFileHandle(self.filepath + '.yaml') as yamlfile:
            yamldata = yaml.safe_load(yamlfile)

        header = yamldata['header']
        sep = yamldata.get('sep', ',')

        dtypes = {}
        columns = []
        for coldata in yamldata['columns']:
            colname = coldata['name']

            dtypes[colname] = coldata['dtype']

            columns.append(colname)

        return header, sep, dtypes, columns

    def generate_metadata(self):
        with getFileHandle(self.filepath) as inputfile:
            header = inputfile.readline().strip()
            sep = self.__detect_sep_from_header(header)
            columns = header.split(sep)
            header = True
            dtypes = self.__generate_dtypes(sep=sep)
            return header, sep, dtypes, columns

    def read_csv(self, chunksize=None, dtypes_override=None, usecols=None):
        dtypes = {k: v for k, v in self.dtypes.items() if v != "NA"}

        if dtypes_override is not None:
            for name, dtype in dtypes_override.items():
                if name in dtypes:
                    dtypes[name] = dtype

        if self.header:
            header = 0
            names = None

        else:
            header = None
            names = self.columns

        dtypes_read = dtypes.copy()
        for column in dtypes_read:
            if dtypes_read[column] == 'int':
                dtypes_read[column] = 'float'
            if dtypes_read[column] == 'bool':
                dtypes_read[column] = 'object'

        try:
            data = pd.read_csv(
                self.filepath, compression=self.compression, chunksize=chunksize,
                sep=self.sep, header=header, names=names, usecols=usecols, dtype=dtypes_read)
        except pd.errors.EmptyDataError:
            data = pd.DataFrame(columns=self.columns)
        except Exception:
           logging.getLogger("single_cell.utils.csv").exception(
               f'loading failed for {self.filepath} with columns {names} and dtypes {dtypes}')
           raise

        def __verify_data(df):
            # Check columns
            cols = self.columns
            if usecols is not None:
                cols = []
                for col in usecols:
                    if col not in self.columns:
                        raise ValueError(f'requested column {col} not in columns')
                for col in self.columns:
                    if col in usecols:
                        cols.append(col)
            if not list(df.columns.values) == cols:
                raise CsvParseError("metadata mismatch in {}".format(self.filepath))

            # recast column data types
            for column in dtypes:
                if usecols is not None and column not in usecols:
                    continue
                try:
                    if dtypes[column] == 'int':
                        df[column] = df[column].astype(int)
                    elif dtypes[column] == 'bool':
                        df.loc[df[column] == 'True', column] = True
                        df.loc[df[column] == 'False', column] = False
                        values = set(df[column].unique())
                        if len(values - {False, True, np.nan}) != 0:
                            raise ValueError(f'{column} is expected to be bool, has values {values}')
                except ValueError:
                    logging.getLogger("single_cell.utils.csv").error(
                        f'unable to convert column {column} to {dtypes[column]}')

        def return_gen(df_iterator):
            for df in df_iterator:
                __verify_data(df)
                yield df

        if chunksize:
            return return_gen(data)
        else:
            __verify_data(data)
            return data


class CsvOutput(object):
    def __init__(
            self, filepath, header=True, sep=',', columns=None, dtypes=None,
            na_rep='NA'
    ):
        self.filepath = filepath
        self.header = header
        self.columns = columns
        self.sep = sep
        self.dtypes = dtypes if dtypes else {}
        self.na_rep = na_rep

    @property
    def yaml_file(self):
        return self.filepath + '.yaml'

    @property
    def header_line(self):
        return self.sep.join(self.columns) + '\n'

    def __get_compression_type_pandas(self):
        filepath = self.filepath
        if filepath.endswith('.tmp'):
            filepath = filepath[:-4]

        _, ext = os.path.splitext(filepath)

        if ext == ".csv":
            return None
        elif ext == ".gz":
            return "gzip"
        elif ext == ".h5" or ext == ".hdf5":
            raise CsvInputError("HDF is not supported")
        else:
            logging.getLogger("single_cell.utils.csv").warn(
                "Couldn't detect output format for {}. extension {}".format(filepath, ext)
            )
            return None

    def __write_yaml(self):

        yamldata = {'header': self.header, 'sep': self.sep, 'columns': []}

        for column in self.columns:
            data = {'name': column, 'dtype': self.dtypes[column]}
            yamldata['columns'].append(data)

        with getFileHandle(self.yaml_file, 'wt') as f:
            yaml.safe_dump(yamldata, f, default_flow_style=False)

    def write_df(self, df):
        if self.columns:
            if not self.columns == df.columns.values:
                raise CsvWriterError("Writer initialized with wrong col names")

        header = df.columns.values if self.header else False
        compression = self.__get_compression_type_pandas()
        df.to_csv(
            self.filepath, sep=self.sep, na_rep=self.na_rep,
            index=False, header=header, compression=compression
        )

        self.columns = list(df.columns.values)
        self.dtypes = get_dtypes_from_df(df)

        self.__write_yaml()

    def write_csv_data(self, reader, writer):
        reader_gzip = type(reader) == gzip.GzipFile
        writer_gzip = type(writer) == gzip.GzipFile

        if reader_gzip == writer_gzip:
            same_format = True
        else:
            same_format = False

        if same_format:
            shutil.copyfileobj(
                reader, writer, length=16 * 1024 * 1024
            )
        else:
            for line in reader:
                writer.write(line)


    def concatenate_files(self, infiles):
        header = self.header_line if self.header else None

        with getFileHandle(self.filepath, 'wt') as writer:
            if header:
                writer.write(header)
            for infile in infiles:
                with getFileHandle(infile) as reader:
                    self.write_csv_data(reader, writer)

        self.__write_yaml()

    def write_headerless_csv(self, infile):
        with getFileHandle(self.filepath, 'wt') as writer:
            with getFileHandle(infile) as reader:
                if not reader.readline() == self.header_line:
                    raise CsvWriterError("cannot write, wrong header")
                self.write_csv_data(reader, writer)

        self.__write_yaml()

    def write_csv_with_header(self, infile):
        with getFileHandle(self.filepath, 'wt') as writer:
            with getFileHandle(infile) as reader:
                writer.write(self.header_line)
                self.write_csv_data(reader, writer)

        self.__write_yaml()


def annotate_csv(infile, annotation_data, outfile, on="cell_id", write_header=True):
    csvinput = CsvInput(infile)

    metrics_df = csvinput.read_csv()

    merge_on = metrics_df[on]

    for cell in merge_on:
        col_data = annotation_data[cell]

        for column, value in col_data.items():
            metrics_df.loc[metrics_df[on] == cell, column] = value

    output = CsvOutput(outfile, sep=csvinput.sep, header=write_header)
    output.write_df(metrics_df)


def concatenate_csv(in_filenames, out_filename, key_column=None, write_header=True):
    if not isinstance(in_filenames, dict):
        in_filenames = dict(enumerate(in_filenames))

    data = []
    sep = None

    for key, in_filename in in_filenames.items():
        csvinput = CsvInput(in_filename)

        if not sep:
            sep = csvinput.sep
        assert sep == csvinput.sep

        df = csvinput.read_csv()

        if key_column is not None:
            df[key_column] = str(key)
        data.append(df)
    data = pd.concat(data, ignore_index=True)

    csvoutput = CsvOutput(out_filename, header=write_header, sep=sep)
    csvoutput.write_df(data)


def extrapolate_types_from_yaml_files(csv_files):
    precedence = ['str', 'float', 'int', 'bool', 'NA']

    csv_metadata = [CsvInput(csv_file) for csv_file in csv_files]

    header = set([val.header for val in csv_metadata])
    assert len(header) == 1, 'mismatched yaml files'
    header = list(header)[0]

    sep = set([val.sep for val in csv_metadata])
    assert len(sep) == 1, 'mismatched yaml files'
    sep = list(sep)[0]

    cols = set([tuple(val.columns) for val in csv_metadata])
    assert len(cols) == 1, 'mismatched yaml files'
    cols = list(cols)[0]

    dtypes = [val.dtypes for val in csv_metadata]

    merged_dtypes = {}
    for dtype_val in dtypes:
        for col, dtype in dtype_val.items():
            if col not in merged_dtypes:
                merged_dtypes[col] = dtype
            else:
                og_index = precedence.index(merged_dtypes[col])
                new_index = precedence.index(dtype)
                if new_index < og_index:
                    merged_dtypes[col] = dtype

    return header, sep, cols, merged_dtypes


def concatenate_csv_files_quick_lowmem(inputfiles, output, write_header=True):
    if isinstance(inputfiles, dict):
        inputfiles = inputfiles.values()

    header, sep, cols, dtypes = extrapolate_types_from_yaml_files(inputfiles)

    if header:
        raise CsvInputError("Attempting to concatenate files with header.")

    csvoutput = CsvOutput(
        output, header=write_header, sep=sep, columns=cols, dtypes=dtypes
    )
    csvoutput.concatenate_files(inputfiles)


def prep_csv_files(filepath, outputfile, header=False):
    """
    generate header less csv files
    :param filepath:
    :type filepath:
    :param outputfile:
    :type outputfile:
    """
    csvinput = CsvInput(filepath)

    if csvinput.header is None:
        raise CsvInputError("cannot prep csv file with no header")

    csvoutput = CsvOutput(
        outputfile, header=False, columns=csvinput.columns,
        sep=csvinput.sep, dtypes=csvinput.dtypes
    )

    if header:
        csvoutput.write_csv_with_header(filepath)
    else:
        csvoutput.write_headerless_csv(filepath)


def finalize_csv(infile, outfile):
    csvinput = CsvInput(infile)

    if csvinput.header:
        raise CsvInputError("cannot finalize file with header")

    csvoutput = CsvOutput(
        outfile, header=True, columns=csvinput.columns,
        sep=csvinput.sep, dtypes=csvinput.dtypes
    )
    csvoutput.write_csv_with_header(infile)


def merge_csv(in_filenames, out_filename, how, on, nan_val='NA', suffixes=None, write_header=True):
    if isinstance(in_filenames, dict):
        in_filenames = in_filenames.values()

    data = []
    sep = None
    for in_filename in in_filenames:
        csvinput = CsvInput(in_filename)
        indata = csvinput.read_csv()
        if not indata.empty:
            data.append(indata)

        if not sep:
            sep = csvinput.sep
        assert sep == csvinput.sep

    data = merge_frames(data, how, on, suffixes=suffixes)
    data = data.fillna(nan_val)

    csvoutput = CsvOutput(out_filename, header=write_header, sep=sep)
    csvoutput.write_df(data)


def merge_frames(frames, how, on, suffixes=None):
    '''
    annotates input_df using ref_df
    '''

    suff = ['', '']

    if ',' in on:
        on = on.split(',')

    if len(frames) == 1:
        return frames[0]
    else:
        left = frames[0]
        right = frames[1]

        cols_to_use = list(right.columns.difference(left.columns))
        cols_to_use += on
        cols_to_use = list(set(cols_to_use))

        if suffixes:
            suff = (suffixes[0], suffixes[1])

        merged_frame = pd.merge(left, right[cols_to_use],
                                how=how,
                                on=on,
                                suffixes=suff)
        for i, frame in enumerate(frames[2:]):

            if suffixes:
                suff = (suffixes[i + 2], suffixes[i + 2])

            merged_frame = pd.merge(merged_frame, frame,
                                    how=how,
                                    on=on,
                                    suffixes=suff)
        return merged_frame


def read_csv_and_yaml(infile, chunksize=None):
    return CsvInput(infile).read_csv(chunksize=chunksize)


def write_dataframe_to_csv_and_yaml(df, outfile, write_header=False, sep=','):
    csvoutput = CsvOutput(outfile, header=write_header, sep=sep)
    csvoutput.write_df(df)

def get_metadata(infile):
    csvinput = CsvInput(infile)
    return csvinput.header, csvinput.dtypes, csvinput.columns



from numba import jit
import numpy as np
import pandas as pd


@jit(nopython=True)
def count_pairs(group_items):
    num_items = group_items.shape[1]
    pair_counts = np.zeros((num_items, num_items))
    for group_idx in range(group_items.shape[0]):
        items = np.where(group_items[group_idx] == 1)[0]
        for item_idx1 in items:
            for item_idx2 in items:
                pair_counts[item_idx1, item_idx2] += 1
    return pair_counts


def generate_cn_transitions(cn_data):
    """ Generate a list of copy number transition regions per cell.

    Args:
        cn_data (DataFrame): copy number data
    
    Returns:
        DataFrame: copy number transitions per cell

    """

    # Build a list of transition bins
    cn_transitions = cn_data.sort_values(['cell_id', 'chr', 'start'])[['cell_id', 'chr', 'start', 'end', 'state']]
    cn_transitions['state_diff'] = cn_transitions['state'].diff().fillna(0).astype(int)
    cn_transitions['chr_diff'] = (cn_transitions['chr'].shift(1) == cn_transitions['chr']) * 1
    cn_transitions['transition'] = (cn_transitions['state_diff'] != 0) & (cn_transitions['chr_diff'] == 1)
    cn_transitions['start'] -= int(500000/2)
    cn_transitions['end'] -= int(500000/2)
    cn_transitions['orientation'] = '+'
    cn_transitions.loc[cn_transitions['state_diff'] > 0, 'orientation'] = '-'
    cn_transitions = cn_transitions.query('transition')
    cn_transitions = cn_transitions.merge(
        cn_transitions[['chr', 'start', 'end', 'orientation']]
            .drop_duplicates()
            .reset_index(drop=True)
            .reset_index()
            .rename(columns={'index': 'region_index'}))

    return cn_transitions


def generate_cn_transition_counts(cn_data):
    """ Generate a list of copy number transition regions and their cell counts.

    Args:
        cn_data (DataFrame): copy number data
    
    Returns:
        DataFrame: copy number transitions cell counts

    """

    cn_transitions = generate_cn_transitions(cn_data)

    # Build a table of transition cell counts
    cn_transition_counts = (
        cn_transitions
            .groupby(['chr', 'start', 'end', 'region_index', 'orientation'], observed=True)
            .size().rename('cell_count').reset_index())

    return cn_transition_counts


def generate_cn_transition_pair_counts(cn_data):
    """ Generate a list of copy number transition pairs and their cell counts.

    Args:
        cn_data (DataFrame): copy number data
    
    Returns:
        DataFrame: copy number transitions pairs and counts of supporting cells

    """

    cn_transitions = generate_cn_transitions(cn_data)

    # Build a matrix of pairs of transitions in the same cell
    cell_transitions_matrix = pd.crosstab(
        index=cn_transitions['cell_id'],
        columns=cn_transitions['region_index'],
    )

    # Possibly not necessary, but ensure the column ordering matches the column indexing
    cell_transitions_matrix = cell_transitions_matrix.reindex(
        columns=range(max(cell_transitions_matrix.columns) + 1))

    # For each pair of transitions, count the number of cells with both
    # returned as a transition by transition matrix
    cell_region_counts = count_pairs(cell_transitions_matrix.values)

    # Create a stacked format matrix that does not include 0 entries
    cell_region_counts[cell_region_counts == 0] = np.NaN

    cell_region_counts = pd.DataFrame(
        cell_region_counts,
        columns=cell_transitions_matrix.columns,
        index=cell_transitions_matrix.columns,
    )

    cn_transitions_matrix = (
        cell_region_counts
            .stack().rename('pair_cell_count')
            .rename_axis(index=['region_index_1', 'region_index_2'])
            .reset_index())

    cn_transitions_matrix['pair_cell_count'] = cn_transitions_matrix['pair_cell_count'].astype(int)

    return cn_transitions_matrix
